"""
NFTSocket -- low level nftables API

See also: pyroute2.nftables
"""

import struct
import threading

from pyroute2.netlink import (
    NETLINK_NETFILTER,
    NLM_F_ACK,
    NLM_F_APPEND,
    NLM_F_CREATE,
    NLM_F_DUMP,
    NLM_F_EXCL,
    NLM_F_REPLACE,
    NLM_F_REQUEST,
    nla,
    nla_base_string,
    nlmsg_atoms,
)
from pyroute2.netlink.nfnetlink import NFNL_SUBSYS_NFTABLES, nfgen_msg
from pyroute2.netlink.nlsocket import NetlinkSocket

NFT_MSG_NEWTABLE = 0
NFT_MSG_GETTABLE = 1
NFT_MSG_DELTABLE = 2
NFT_MSG_NEWCHAIN = 3
NFT_MSG_GETCHAIN = 4
NFT_MSG_DELCHAIN = 5
NFT_MSG_NEWRULE = 6
NFT_MSG_GETRULE = 7
NFT_MSG_DELRULE = 8
NFT_MSG_NEWSET = 9
NFT_MSG_GETSET = 10
NFT_MSG_DELSET = 11
NFT_MSG_NEWSETELEM = 12
NFT_MSG_GETSETELEM = 13
NFT_MSG_DELSETELEM = 14
NFT_MSG_NEWGEN = 15
NFT_MSG_GETGEN = 16
NFT_MSG_TRACE = 17
NFT_MSG_NEWOBJ = 18
NFT_MSG_GETOBJ = 19
NFT_MSG_DELOBJ = 20
NFT_MSG_GETOBJ_RESET = 21
NFT_MSG_NEWFLOWTABLE = 22
NFT_MSG_GETFLOWTABLE = 23
NFT_MSG_DELFLOWTABLE = 24

# from nftables/include/datatype.h
DATA_TYPE_INVALID = 0
DATA_TYPE_VERDICT = 1
DATA_TYPE_NFPROTO = 2
DATA_TYPE_BITMASK = 3
DATA_TYPE_INTEGER = 4
DATA_TYPE_STRING = 5
DATA_TYPE_LLADDR = 6
DATA_TYPE_IPADDR = 7
DATA_TYPE_IP6ADDR = 8
DATA_TYPE_ETHERADDR = 9
DATA_TYPE_ETHERTYPE = 10
DATA_TYPE_ARPOP = 11
DATA_TYPE_INET_PROTOCOL = 12
DATA_TYPE_INET_SERVICE = 13
DATA_TYPE_ICMP_TYPE = 14
DATA_TYPE_TCP_FLAG = 15
DATA_TYPE_DCCP_PKTTYPE = 16
DATA_TYPE_MH_TYPE = 17
DATA_TYPE_TIME = 18
DATA_TYPE_MARK = 19
DATA_TYPE_IFINDEX = 20
DATA_TYPE_ARPHRD = 21
DATA_TYPE_REALM = 22
DATA_TYPE_CLASSID = 23
DATA_TYPE_UID = 24
DATA_TYPE_GID = 25
DATA_TYPE_CT_STATE = 26
DATA_TYPE_CT_DIR = 27
DATA_TYPE_CT_STATUS = 28
DATA_TYPE_ICMP6_TYPE = 29
DATA_TYPE_CT_LABEL = 30
DATA_TYPE_PKTTYPE = 31
DATA_TYPE_ICMP_CODE = 32
DATA_TYPE_ICMPV6_CODE = 33
DATA_TYPE_ICMPX_CODE = 34
DATA_TYPE_DEVGROUP = 35
DATA_TYPE_DSCP = 36
DATA_TYPE_ECN = 37
DATA_TYPE_FIB_ADDR = 38
DATA_TYPE_BOOLEAN = 39
DATA_TYPE_CT_EVENTBIT = 40
DATA_TYPE_IFNAME = 41
DATA_TYPE_IGMP_TYPE = 42
DATA_TYPE_TIME_DATE = 43
DATA_TYPE_TIME_HOUR = 44
DATA_TYPE_TIME_DAY = 45
DATA_TYPE_CGROUPV2 = 46

# from include/uapi/linux/netfilter.h
NFPROTO_INET = 1
NFPROTO_IPV4 = 2
NFPROTO_ARP = 3
NFPROTO_NETDEV = 5
NFPROTO_BRIDGE = 7
NFPROTO_IPV6 = 10


class nftnl_udata(nla_base_string):
    # TLV structures:
    # nftnl_udata
    #  <-------- HEADER --------> <------ PAYLOAD ------>
    # +------------+-------------+- - - - - - - - - - - -+
    # |    type    |     len     |         value         |
    # |  (1 byte)  |   (1 byte)  |                       |
    # +--------------------------+- - - - - - - - - - - -+
    #  <-- sizeof(nftnl_udata) -> <-- nftnl_udata->len -->
    __slots__ = ()

    @classmethod
    def udata_decode(cls, data):
        offset = 0
        result = []
        while offset + 2 < len(data):
            utype = data[offset]
            ulen = data[offset + 1]
            offset += 2
            if offset + ulen > len(data):
                return None  # bad decode
            try:
                type_name = cls.udata_types[utype]
            except IndexError:
                return None  # bad decode

            value = data[offset : offset + ulen]
            if type_name.endswith("_COMMENT") and value[-1] == 0:
                value = value[:-1]  # remove \x00 c *str
            result.append((type_name, value))
            offset += ulen
        return result

    @classmethod
    def udata_encode(cls, values):
        value = b""
        for type_name, udata in values:
            if isinstance(udata, str):
                udata = udata.encode()
            if type_name.endswith("_COMMENT") and udata[-1] != 0:
                udata = udata + b"\x00"
            utype = cls.udata_types.index(type_name)
            value += struct.pack("BB", utype, len(udata)) + udata
        return value

    def decode(self):
        nla_base_string.decode(self)
        value = self.udata_decode(self['value'])
        if value is not None:
            self.value = value

    def encode(self):
        if not isinstance(self.value, (bytes, str)):
            self['value'] = self.udata_encode(self.value)
        nla_base_string.encode(self)


class nft_map_uint8(nla):
    ops = {}
    fields = [('value', 'B')]

    def decode(self):
        nla.decode(self)
        self.value = self.ops.get(self['value'])


class nft_map_be32(nft_map_uint8):
    fields = [('value', '>I')]


class nft_map_be32_signed(nft_map_uint8):
    fields = [('value', '>i')]


class nft_flags_be32(nla):
    fields = [('value', '>I')]
    ops = None

    def decode(self):
        nla.decode(self)
        self.value = frozenset(
            o for i, o in enumerate(self.ops) if self['value'] & 1 << i
        )

    def encode(self):
        value = 0
        for i, name in enumerate(self.ops):
            if name in self.value:
                value |= 1 << i
        self["value"] = value
        nla.encode(self)


class nft_flags_be16(nla):
    fields = [('value', '>H')]
    ops = None

    def decode(self):
        nla.decode(self)
        self.value = frozenset(
            o for i, o in enumerate(self.ops) if self['value'] & 1 << i
        )


class nft_device(nla):
    class device_attributes(nla):
        nla_map = (
            ('NFTA_DEVICE_UNSPEC', 'none'),
            ('NFTA_DEVICE_NAME', 'asciiz'),
        )


class nft_gen_msg(nfgen_msg):
    nla_map = (
        ('NFTA_GEN_UNSPEC', 'none'),
        ('NFTA_GEN_ID', 'be32'),
        ('NFTA_GEN_PROC_PID', 'be32'),
        ('NFTA_GEN_PROC_NAME', 'asciiz'),
    )


class nft_chain_msg(nfgen_msg):
    prefix = 'NFTA_CHAIN_'
    nla_map = (
        ('NFTA_CHAIN_UNSPEC', 'none'),
        ('NFTA_CHAIN_TABLE', 'asciiz'),
        ('NFTA_CHAIN_HANDLE', 'be64'),
        ('NFTA_CHAIN_NAME', 'asciiz'),
        ('NFTA_CHAIN_HOOK', 'hook'),
        ('NFTA_CHAIN_POLICY', 'be32'),
        ('NFTA_CHAIN_USE', 'be32'),
        ('NFTA_CHAIN_TYPE', 'asciiz'),
        ('NFTA_CHAIN_COUNTERS', 'counters'),
        ('NFTA_CHAIN_PAD', 'hex'),
        ('NFTA_CHAIN_FLAGS', 'flags'),
        ('NFTA_CHAIN_ID', 'be32'),
        ('NFTA_CHAIN_USERDATA', 'hex'),
    )

    class counters(nla):
        nla_map = (
            ('NFTA_COUNTER_UNSPEC', 'none'),
            ('NFTA_COUNTER_BYTES', 'be64'),
            ('NFTA_COUNTER_PACKETS', 'be64'),
        )

    class hook(nft_device):
        nla_map = (
            ('NFTA_HOOK_UNSPEC', 'none'),
            ('NFTA_HOOK_HOOKNUM', 'be32'),
            ('NFTA_HOOK_PRIORITY', 'sbe32'),
            ('NFTA_HOOK_DEV', 'asciiz'),
            ('NFTA_HOOK_DEVS', 'device_attributes'),
        )

    class flags(nft_flags_be32):
        ops = ('NFT_CHAIN_HW_OFFLOAD', 'NFT_CHAIN_BINDING')


class nat_flags(nla):
    class nat_range(nft_flags_be32):
        ops = (
            'NF_NAT_RANGE_MAP_IPS',
            'NF_NAT_RANGE_PROTO_SPECIFIED',
            'NF_NAT_RANGE_PROTO_RANDOM',
            'NF_NAT_RANGE_PERSISTENT',
            'NF_NAT_RANGE_PROTO_RANDOM_FULLY',
            'NF_NAT_RANGE_PROTO_OFFSET',
            'NF_NAT_RANGE_NETMAP',
        )


class nft_regs(nla):
    class regs(nft_map_be32):
        ops = {
            0x00: 'NFT_REG_VERDICT',
            0x01: 'NFT_REG_1',
            0x02: 'NFT_REG_2',
            0x03: 'NFT_REG_3',
            0x04: 'NFT_REG_4',
            0x08: 'NFT_REG32_00',
            0x09: 'NFT_REG32_01',
            0x0A: 'NFT_REG32_02',
            0x0B: 'NFT_REG32_03',
            0x0C: 'NFT_REG32_04',
            0x0D: 'NFT_REG32_05',
            0x0E: 'NFT_REG32_06',
            0x0F: 'NFT_REG32_07',
            0x10: 'NFT_REG32_08',
            0x11: 'NFT_REG32_09',
            0x12: 'NFT_REG32_10',
            0x13: 'NFT_REG32_11',
            0x14: 'NFT_REG32_12',
            0x15: 'NFT_REG32_13',
            0x16: 'NFT_REG32_14',
            0x17: 'NFT_REG32_15',
        }


class nft_data(nla):
    class nfta_data(nla):
        nla_map = (
            ('NFTA_DATA_UNSPEC', 'none'),
            ('NFTA_DATA_VALUE', 'cdata'),
            ('NFTA_DATA_VERDICT', 'verdict'),
        )

        class verdict(nla):
            nla_map = (
                ('NFTA_VERDICT_UNSPEC', 'none'),
                ('NFTA_VERDICT_CODE', 'verdict_code'),
                ('NFTA_VERDICT_CHAIN', 'asciiz'),
            )

            class verdict_code(nft_map_be32_signed):
                ops = {
                    0: 'NF_DROP',
                    1: 'NF_ACCEPT',
                    2: 'NF_STOLEN',
                    3: 'NF_QUEUE',
                    4: 'NF_REPEAT',
                    5: 'NF_STOP',
                    -1: 'NFT_CONTINUE',
                    -2: 'NFT_BREAK',
                    -3: 'NFT_JUMP',
                    -4: 'NFT_GOTO',
                    -5: 'NFT_RETURN',
                }


# nft_expr struct, used for rules and set
class nft_contains_expr:
    class nft_expr(nla):
        header_type = 1

        nla_map = (
            ('NFTA_EXPR_UNSPEC', 'none'),
            ('NFTA_EXPR_NAME', 'asciiz'),
            ('NFTA_EXPR_DATA', 'expr'),
        )

        class nft_bitwise(nft_data, nft_regs):
            nla_map = (
                ('NFTA_BITWISE_UNSPEC', 'none'),
                ('NFTA_BITWISE_SREG', 'regs'),
                ('NFTA_BITWISE_DREG', 'regs'),
                ('NFTA_BITWISE_LEN', 'be32'),
                ('NFTA_BITWISE_MASK', 'nfta_data'),
                ('NFTA_BITWISE_XOR', 'nfta_data'),
                ('NFTA_BITWISE_OP', 'bitwise_op'),
                ('NFTA_BITWISE_DATA', 'nfta_data'),
            )

            class bitwise_op(nft_map_be32):
                ops = {
                    0: 'NFT_BITWISE_BOOL',
                    1: 'NFT_BITWISE_LSHIFT',
                    2: 'NFT_BITWISE_RSHIFT',
                }

        class nft_byteorder(nft_regs):
            nla_map = (
                ('NFTA_BYTEORDER_UNSPEC', 'none'),
                ('NFTA_BYTEORDER_SREG', 'regs'),
                ('NFTA_BYTEORDER_DREG', 'regs'),
                ('NFTA_BYTEORDER_OP', 'ops'),
                ('NFTA_BYTEORDER_LEN', 'be32'),
                ('NFTA_BYTEORDER_SIZE', 'be32'),
            )

            class ops(nft_map_be32):
                ops = {0: 'NFT_BYTEORDER_NTOH', 1: 'NFT_BYTEORDER_HTON'}

        class nft_cmp(nft_data, nft_regs):
            nla_map = (
                ('NFTA_CMP_UNSPEC', 'none'),
                ('NFTA_CMP_SREG', 'regs'),
                ('NFTA_CMP_OP', 'ops'),
                ('NFTA_CMP_DATA', 'nfta_data'),
            )

            class ops(nft_map_be32):
                ops = {
                    0: 'NFT_CMP_EQ',
                    1: 'NFT_CMP_NEQ',
                    2: 'NFT_CMP_LT',
                    3: 'NFT_CMP_LTE',
                    4: 'NFT_CMP_GT',
                    5: 'NFT_CMP_GTE',
                }

        class nft_match(nla):
            nla_map = (
                ('NFTA_MATCH_UNSPEC', 'none'),
                ('NFTA_MATCH_NAME', 'asciiz'),
                ('NFTA_MATCH_REV', 'be32'),
                ('NFTA_MATCH_INFO', 'hex'),
                ('NFTA_MATCH_PROTOCOL', 'hex'),
                ('NFTA_MATCH_FLAGS', 'hex'),
            )

        class nft_target(nla):
            nla_map = (
                ('NFTA_TARGET_UNSPEC', 'none'),
                ('NFTA_TARGET_NAME', 'asciiz'),
                ('NFTA_TARGET_REV', 'be32'),
                ('NFTA_TARGET_INFO', 'hex'),
                ('NFTA_TARGET_PROTOCOL', 'hex'),
                ('NFTA_TARGET_FLAGS', 'hex'),
            )

        class nft_connlimit(nla):
            nla_map = (
                ('NFTA_CONNLIMIT_UNSPEC', 'none'),
                ('NFTA_CONNLIMIT_COUNT', 'be32'),
                ('NFTA_CONNLIMIT_FLAGS', 'connlimit_flags'),
            )

            class connlimit_flags(nft_flags_be32):
                ops = ('NFT_LIMIT_F_INV',)

        class nft_counter(nla):
            nla_map = (
                ('NFTA_COUNTER_UNSPEC', 'none'),
                ('NFTA_COUNTER_BYTES', 'be64'),
                ('NFTA_COUNTER_PACKETS', 'be64'),
            )

        class nft_ct(nft_regs):
            nla_map = (
                ('NFTA_CT_UNSPEC', 'none'),
                ('NFTA_CT_DREG', 'regs'),
                ('NFTA_CT_KEY', 'keys'),
                ('NFTA_CT_DIRECTION', 'uint8'),
                ('NFTA_CT_SREG', 'regs'),
            )

            class keys(nft_map_be32):
                ops = {
                    0x00: 'NFT_CT_STATE',
                    0x01: 'NFT_CT_DIRECTION',
                    0x02: 'NFT_CT_STATUS',
                    0x03: 'NFT_CT_MARK',
                    0x04: 'NFT_CT_SECMARK',
                    0x05: 'NFT_CT_EXPIRATION',
                    0x06: 'NFT_CT_HELPER',
                    0x07: 'NFT_CT_L3PROTOCOL',
                    0x08: 'NFT_CT_SRC',
                    0x09: 'NFT_CT_DST',
                    0x0A: 'NFT_CT_PROTOCOL',
                    0x0B: 'NFT_CT_PROTO_SRC',
                    0x0C: 'NFT_CT_PROTO_DST',
                    0x0D: 'NFT_CT_LABELS',
                    0x0E: 'NFT_CT_PKTS',
                    0x0F: 'NFT_CT_BYTES',
                    0x10: 'NFT_CT_AVGPKT',
                    0x11: 'NFT_CT_ZONE',
                    0x12: 'NFT_CT_EVENTMASK',
                    0x13: 'NFT_CT_SRC_IP',
                    0x14: 'NFT_CT_DST_IP',
                    0x15: 'NFT_CT_SRC_IP6',
                    0x16: 'NFT_CT_DST_IP6',
                    0x17: 'NFT_CT_ID',
                }

        class nft_dup(nft_regs):
            nla_map = (
                ('NFTA_DUP_UNSPEC', 'none'),
                ('NFTA_DUP_SREG_ADDR', 'regs'),
                ('NFTA_DUP_SREG_DEV', 'regs'),
            )

        class nft_exthdr(nft_regs):
            nla_map = (
                ('NFTA_EXTHDR_UNSPEC', 'none'),
                ('NFTA_EXTHDR_DREG', 'regs'),
                ('NFTA_EXTHDR_TYPE', 'uint8'),
                ('NFTA_EXTHDR_OFFSET', 'be32'),
                ('NFTA_EXTHDR_LEN', 'be32'),
                ('NFTA_EXTHDR_FLAGS', 'exthdr_flags'),
                ('NFTA_EXTHDR_OP', 'exthdr_op'),
                ('NFTA_EXTHDR_SREG', 'regs'),
            )

            class exthdr_flags(nft_flags_be32):
                ops = ('NFT_EXTHDR_F_PRESENT',)

            class exthdr_op(nft_map_be32):
                ops = {
                    0: 'NFT_EXTHDR_OP_IPV6',
                    1: 'NFT_EXTHDR_OP_TCPOPT',
                    2: 'NFT_EXTHDR_OP_IPV4',
                }

        class nft_fib(nft_regs):
            nla_map = (
                ('NFTA_FIB_UNSPEC', 'none'),
                ('NFTA_FIB_DREG', 'regs'),
                ('NFTA_FIB_RESULT', 'fib_result'),
                ('NFTA_FIB_FLAGS', 'fib_flags'),
            )

            class fib_result(nft_flags_be32):
                ops = (
                    'NFT_FIB_RESULT_UNSPEC',
                    'NFT_FIB_RESULT_OIF',
                    'NFT_FIB_RESULT_OIFNAME',
                    'NFT_FIB_RESULT_ADDRTYPE',
                )

            class fib_flags(nft_map_be32):
                ops = {
                    0: 'NFTA_FIB_F_SADDR',
                    1: 'NFTA_FIB_F_DADDR',
                    2: 'NFTA_FIB_F_MARK',
                    3: 'NFTA_FIB_F_IIF',
                    4: 'NFTA_FIB_F_OIF',
                    5: 'NFTA_FIB_F_PRESENT',
                }

        class nft_fwd(nft_regs):
            nla_map = (
                ('NFTA_FWD_UNSPEC', 'none'),
                ('NFTA_FWD_SREG_DEV', 'regs'),
                ('NFTA_FWD_SREG_ADDR', 'regs'),
                ('NFTA_FWD_NFPROTO', 'u32'),
            )

        class nft_hash(nft_regs):
            nla_map = (
                ('NFTA_HASH_UNSPEC', 'none'),
                ('NFTA_HASH_SREG', 'regs'),
                ('NFTA_HASH_DREG', 'regs'),
                ('NFTA_HASH_LEN', 'be32'),
                ('NFTA_HASH_MODULUS', 'be32'),
                ('NFTA_HASH_SEED', 'be32'),
                ('NFTA_HASH_OFFSET', 'be32'),
                ('NFTA_HASH_TYPE', 'hash_type'),
                ('NFTA_HASH_SET_NAME', 'asciiz'),
                ('NFTA_HASH_SET_ID', 'be32'),
            )

            class hash_type(nft_map_be32):
                ops = {0: 'NFT_HASH_JENKINS', 1: 'NFT_HASH_SYM'}

        class nft_immediate(nft_data, nft_regs):
            nla_map = (
                ('NFTA_IMMEDIATE_UNSPEC', 'none'),
                ('NFTA_IMMEDIATE_DREG', 'regs'),
                ('NFTA_IMMEDIATE_DATA', 'nfta_data'),
            )

        class nft_limit(nla):
            nla_map = (
                ('NFTA_LIMIT_UNSPEC', 'none'),
                ('NFTA_LIMIT_RATE', 'be64'),
                ('NFTA_LIMIT_UNIT', 'be64'),
                ('NFTA_LIMIT_BURST', 'be32'),
                ('NFTA_LIMIT_TYPE', 'types'),
                ('NFTA_LIMIT_FLAGS', 'be32'),
            )  # make flags type

            class types(nft_map_be32):
                ops = {0: 'NFT_LIMIT_PKTS', 1: 'NFT_LIMIT_PKT_BYTES'}

        class nft_log(nla):
            nla_map = (
                ('NFTA_LOG_UNSPEC', 'none'),
                ('NFTA_LOG_GROUP', 'be32'),
                ('NFTA_LOG_PREFIX', 'asciiz'),
                ('NFTA_LOG_SNAPLEN', 'be32'),
                ('NFTA_LOG_QTHRESHOLD', 'be32'),
                ('NFTA_LOG_LEVEL', 'log_level'),
                ('NFTA_LOG_FLAGS', 'log_flags'),
            )

            class log_level(nft_map_be32):
                ops = {
                    0: 'NFT_LOGLEVEL_EMERG',
                    1: 'NFT_LOGLEVEL_ALERT',
                    2: 'NFT_LOGLEVEL_CRIT',
                    3: 'NFT_LOGLEVEL_ERR',
                    4: 'NFT_LOGLEVEL_WARNING',
                    5: 'NFT_LOGLEVEL_NOTICE',
                    6: 'NFT_LOGLEVEL_INFO',
                    7: 'NFT_LOGLEVEL_DEBUG',
                    8: 'NFT_LOGLEVEL_AUDIT',
                }

            class log_flags(nft_flags_be32):
                ops = (
                    'NF_LOG_TCPSEQ',
                    'NF_LOG_TCPOPT',
                    'NF_LOG_IPOPT',
                    'NF_LOG_UID',
                    'NF_LOG_NFLOG',
                    'NF_LOG_MACDECODE',
                )

        class nft_lookup(nft_regs):
            nla_map = (
                ('NFTA_LOOKUP_UNSPEC', 'none'),
                ('NFTA_LOOKUP_SET', 'asciiz'),
                ('NFTA_LOOKUP_SREG', 'regs'),
                ('NFTA_LOOKUP_DREG', 'regs'),
                ('NFTA_LOOKUP_SET_ID', 'be32'),
                ('NFTA_LOOKUP_FLAGS', 'lookup_flags'),
            )

            class lookup_flags(nft_flags_be32):
                ops = ('NFT_LOOKUP_F_INV',)

        class nft_masq(nft_regs, nat_flags):
            nla_map = (
                ('NFTA_MASQ_UNSPEC', 'none'),
                ('NFTA_MASQ_FLAGS', 'nat_range'),
                ('NFTA_MASQ_REG_PROTO_MIN', 'regs'),
                ('NFTA_MASQ_REG_PROTO_MAX', 'regs'),
            )

        class nft_meta(nft_regs):
            nla_map = (
                ('NFTA_META_UNSPEC', 'none'),
                ('NFTA_META_DREG', 'regs'),
                ('NFTA_META_KEY', 'meta_key'),
                ('NFTA_META_SREG', 'regs'),
            )

            class meta_key(nft_map_be32):
                ops = {
                    0: 'NFT_META_LEN',
                    1: 'NFT_META_PROTOCOL',
                    2: 'NFT_META_PRIORITY',
                    3: 'NFT_META_MARK',
                    4: 'NFT_META_IIF',
                    5: 'NFT_META_OIF',
                    6: 'NFT_META_IIFNAME',
                    7: 'NFT_META_OIFNAME',
                    8: 'NFT_META_IIFTYPE',
                    9: 'NFT_META_OIFTYPE',
                    10: 'NFT_META_SKUID',
                    11: 'NFT_META_SKGID',
                    12: 'NFT_META_NFTRACE',
                    13: 'NFT_META_RTCLASSID',
                    14: 'NFT_META_SECMARK',
                    15: 'NFT_META_NFPROTO',
                    16: 'NFT_META_L4PROTO',
                    17: 'NFT_META_BRI_IIFNAME',
                    18: 'NFT_META_BRI_OIFNAME',
                    19: 'NFT_META_PKTTYPE',
                    20: 'NFT_META_CPU',
                    21: 'NFT_META_IIFGROUP',
                    22: 'NFT_META_OIFGROUP',
                    23: 'NFT_META_CGROUP',
                    24: 'NFT_META_PRANDOM',
                    25: 'NFT_META_SECPATH',
                    26: 'NFT_META_IIFKIND',
                    27: 'NFT_META_OIFKIND',
                    28: 'NFT_META_BRI_IIFPVID',
                    29: 'NFT_META_BRI_IIFVPROTO',
                    30: 'NFT_META_TIME_NS',
                    31: 'NFT_META_TIME_DAY',
                    32: 'NFT_META_TIME_HOUR',
                    33: 'NFT_META_SDIF',
                    34: 'NFT_META_SDIFNAME',
                }

        class nft_nat(nft_regs, nat_flags):
            nla_map = (
                ('NFTA_NAT_UNSPEC', 'none'),
                ('NFTA_NAT_TYPE', 'types'),
                ('NFTA_NAT_FAMILY', 'be32'),
                ('NFTA_NAT_REG_ADDR_MIN', 'regs'),
                ('NFTA_NAT_REG_ADDR_MAX', 'regs'),
                ('NFTA_NAT_REG_PROTO_MIN', 'regs'),
                ('NFTA_NAT_REG_PROTO_MAX', 'regs'),
                ('NFTA_NAT_FLAGS', 'nat_range'),
            )

            class types(nft_map_be32):
                ops = {0: 'NFT_NAT_SNAT', 1: 'NFT_NAT_DNAT'}

        class nft_numgen(nft_regs):
            nla_map = (
                ('NFTA_NG_UNSPEC', 'none'),
                ('NFTA_NG_DREG', 'regs'),
                ('NFTA_NG_MODULUS', 'be32'),
                ('NFTA_NG_TYPE', 'types'),
                ('NFTA_NG_OFFSET', 'be32'),
                ('NFTA_NG_SET_NAME', 'asciiz'),
                ('NFTA_NG_SET_ID', 'be32'),
            )

            class types(nft_map_be32):
                ops = {0: 'NFT_NG_INCREMENTAL', 1: 'NFT_NG_RANDOM'}

        class nft_objref(nft_regs):
            nla_map = (
                ('NFTA_OBJREF_UNSPEC', 'none'),
                ('NFTA_OBJREF_IMM_TYPE', 'regs'),
                ('NFTA_OBJREF_IMM_NAME', 'asciiz'),
                ('NFTA_OBJREF_SET_SREG', 'regs'),
                ('NFTA_OBJREF_SET_NAME', 'asciiz'),
                ('NFTA_OBJREF_SET_ID', 'be32'),
            )

        class nft_offload(nla):
            nla_map = (
                ('NFTA_FLOW_UNSPEC', 'none'),
                ('NFTA_FLOW_TABLE_NAME', 'asciiz'),
            )

        class nft_osf(nft_regs):
            nla_map = (
                ('NFTA_OSF_UNSPEC', 'none'),
                ('NFTA_OSF_DREG', 'regs'),
                ('NFTA_OSF_TTL', 'uint8'),
                ('NFTA_OSF_FLAGS', 'osf_flags'),
            )

            class osf_flags(nft_flags_be32):
                ops = ('NFT_OSF_F_VERSION',)

        class nft_payload(nft_regs):
            nla_map = (
                ('NFTA_PAYLOAD_UNSPEC', 'none'),
                ('NFTA_PAYLOAD_DREG', 'regs'),
                ('NFTA_PAYLOAD_BASE', 'base_type'),
                ('NFTA_PAYLOAD_OFFSET', 'be32'),
                ('NFTA_PAYLOAD_LEN', 'be32'),
                ('NFTA_PAYLOAD_SREG', 'regs'),
                ('NFTA_PAYLOAD_CSUM_TYPE', 'csum_type'),
                ('NFTA_PAYLOAD_CSUM_OFFSET', 'be32'),
                ('NFTA_PAYLOAD_CSUM_FLAGS', 'csum_flags'),
            )

            class base_type(nft_map_be32):
                ops = {
                    0: 'NFT_PAYLOAD_LL_HEADER',
                    1: 'NFT_PAYLOAD_NETWORK_HEADER',
                    2: 'NFT_PAYLOAD_TRANSPORT_HEADER',
                }

            class csum_type(nft_map_be32):
                ops = {
                    0: 'NFT_PAYLOAD_CSUM_NONE',
                    1: 'NFT_PAYLOAD_CSUM_INET',  # RFC 791
                    2: 'NFT_PAYLOAD_CSUM_SCTP',
                }  # RFC 3309

            class csum_flags(nft_flags_be32):
                ops = ('NFT_PAYLOAD_L4CSUM_PSEUDOHDR',)

        class nft_queue(nft_regs):
            nla_map = (
                ('NFTA_QUEUE_UNSPEC', 'none'),
                ('NFTA_QUEUE_NUM', 'be16'),
                ('NFTA_QUEUE_TOTAL', 'be16'),
                ('NFTA_QUEUE_FLAGS', 'queue_flags'),
                ('NFTA_QUEUE_SREG_QNUM', 'regs'),
            )

            class queue_flags(nft_flags_be16):
                ops = ('NFT_QUEUE_FLAG_BYPASS', 'NFT_QUEUE_FLAG_CPU_FANOUT')

        class nft_quota(nla):
            nla_map = (
                ('NFTA_QUOTA_UNSPEC', 'none'),
                ('NFTA_QUOTA_BYTES', 'be16'),
                ('NFTA_QUOTA_FLAGS', 'quota_flags'),
                ('NFTA_QUOTA_PAD', 'hex'),
                ('NFTA_QUOTA_CONSUMED', 'be64'),
            )

            class quota_flags(nft_flags_be32):
                ops = ('NFT_QUOTA_F_INV', 'NFT_QUOTA_F_DEPLETED')

        class nft_range(nft_regs, nft_data):
            nla_map = (
                ('NFTA_RANGE_UNSPEC', 'none'),
                ('NFTA_RANGE_SREG', 'regs'),
                ('NFTA_RANGE_OP', 'range_op'),
                ('NFTA_RANGE_FROM_DATA', 'nfta_data'),
                ('NFTA_RANGE_TO_DATA', 'nfta_data'),
            )

            class range_op(nft_map_be32):
                ops = {0: 'NFT_RANGE_EQ', 1: 'NFT_RANGE_NEQ'}

        class nft_redir(nft_regs, nat_flags):
            nla_map = (
                ('NFTA_REDIR_UNSPEC', 'none'),
                ('NFTA_REDIR_REG_PROTO_MIN', 'regs'),
                ('NFTA_REDIR_REG_PROTO_MAX', 'regs'),
                ('NFTA_REDIR_FLAGS', 'nat_range'),
            )

        class nft_reject(nla):
            nla_map = (
                ('NFTA_REJECT_UNSPEC', 'none'),
                ('NFTA_REJECT_TYPE', 'types'),
                ('NFTA_REJECT_ICMP_CODE', 'codes'),
            )

            class types(nft_map_be32):
                ops = {
                    0: 'NFT_REJECT_ICMP_UNREACH',
                    1: 'NFT_REJECT_TCP_RST',
                    2: 'NFT_REJECT_ICMPX_UNREACH',
                }

            class codes(nft_map_uint8):
                ops = {
                    0: 'NFT_REJECT_ICMPX_NO_ROUTE',
                    1: 'NFT_REJECT_ICMPX_PORT_UNREACH',
                    2: 'NFT_REJECT_ICMPX_HOST_UNREACH',
                    3: 'NFT_REJECT_ICMPX_ADMIN_PROHIBITED',
                }

        class nft_rt(nft_regs):
            nla_map = (
                ('NFTA_RT_UNSPEC', 'none'),
                ('NFTA_RT_DREG', 'regs'),
                ('NFTA_RT_KEY', 'rt_keys'),
            )

            class rt_keys(nft_map_be32):
                ops = {
                    0: 'NFT_RT_CLASSID',
                    1: 'NFT_RT_NEXTHOP4',
                    2: 'NFT_RT_NEXTHOP6',
                    3: 'NFT_RT_TCPMSS',
                    4: 'NFT_RT_XFRM',
                }

        class nft_secmark(nla):
            nla_map = (
                ('NFTA_SECMARK_UNSPEC', 'none'),
                ('NFTA_SECMARK_CTX', 'asciiz'),
            )

        class nft_socket(nft_regs):
            nla_map = (
                ('NFTA_SOCKET_UNSPEC', 'none'),
                ('NFTA_SOCKET_KEY', 'socket_keys'),
                ('NFTA_SOCKET_DREG', 'regs'),
            )

            class socket_keys(nft_map_be32):
                ops = {
                    0: 'NFT_SOCKET_TRANSPARENT',
                    1: 'NFT_SOCKET_MARK',
                    2: 'NFT_SOCKET_WILDCARD',
                }

        class nft_synproxy(nla):
            nla_map = (
                ('NFTA_SYNPROXY_UNSPEC', 'none'),
                ('NFTA_SYNPROXY_MSS', 'u16'),
                ('NFTA_SYNPROXY_WSCALE', 'uint8'),
                ('NFTA_SYNPROXY_FLAGS', 'synproxy_flags'),
            )

            class synproxy_flags(nft_flags_be32):
                ops = (
                    'NF_SYNPROXY_OPT_MSS',
                    'NF_SYNPROXY_OPT_WSCALE',
                    'NF_SYNPROXY_OPT_SACK_PERM',
                    'NF_SYNPROXY_OPT_TIMESTAMP',
                    'NF_SYNPROXY_OPT_ECN',
                )

        class nft_tproxy(nft_regs):
            nla_map = (
                ('NFTA_TPROXY_UNSPEC', 'none'),
                ('NFTA_TPROXY_FAMILY', 'regs'),
                ('NFTA_TPROXY_REG_ADDR', 'regs'),
                ('NFTA_TPROXY_REG_PORT', 'regs'),
            )

        class nft_dynset(nft_regs):
            rule_expr = None
            nla_map = (
                ('NFTA_DYNSET_UNSPEC', 'none'),
                ('NFTA_DYNSET_SET_NAME', 'asciiz'),
                ('NFTA_DYNSET_SET_ID', 'be32'),
                ('NFTA_DYNSET_OP', 'dynset_op'),
                ('NFTA_DYNSET_SREG_KEY', 'regs'),
                ('NFTA_DYNSET_SREG_DATA', 'regs'),
                ('NFTA_DYNSET_TIMEOUT', 'be64'),
                ('NFTA_DYNSET_EXPR', 'rule_expr'),
                ('NFTA_DYNSET_PAD', 'hex'),
                ('NFTA_DYNSET_FLAGS', 'dynset_flags'),
            )

            class dynset_flags(nft_flags_be32):
                ops = ('NFT_DYNSET_F_INV',)

            class dynset_op(nft_map_be32):
                ops = {
                    0: 'NFT_DYNSET_OP_ADD',
                    1: 'NFT_DYNSET_OP_UPDATE',
                    2: 'NFT_DYNSET_OP_DELETE',
                }

        class nft_xfrm(nft_regs):
            nla_map = (
                ('NFTA_XFRM_UNSPEC', 'none'),
                ('NFTA_XFRM_DREG', 'regs'),
                ('NFTA_XFRM_KEY', 'xfrm_key'),
                ('NFTA_XFRM_DIR', 'uint8'),
                ('NFTA_XFRM_SPNUM', 'be32'),
            )

            class xfrm_key(nft_map_be32):
                ops = {
                    0: 'NFT_XFRM_KEY_UNSPEC',
                    1: 'NFT_XFRM_KEY_DADDR_IP4',
                    2: 'NFT_XFRM_KEY_DADDR_IP6',
                    3: 'NFT_XFRM_KEY_SADDR_IP4',
                    4: 'NFT_XFRM_KEY_SADDR_IP6',
                    5: 'NFT_XFRM_KEY_REQID',
                    6: 'NFT_XFRM_KEY_SPI',
                }

        @staticmethod
        def expr(self, *argv, **kwarg):
            data_type = self.get_attr('NFTA_EXPR_NAME')
            expr = getattr(self, 'nft_%s' % data_type, self.hex)
            if hasattr(expr, 'rule_expr'):
                expr.rule_expr = self.__class__
            return expr


class nft_rule_msg(nfgen_msg, nft_contains_expr):
    prefix = 'NFTA_RULE_'
    nla_map = (
        ('NFTA_RULE_UNSPEC', 'none'),
        ('NFTA_RULE_TABLE', 'asciiz'),
        ('NFTA_RULE_CHAIN', 'asciiz'),
        ('NFTA_RULE_HANDLE', 'be64'),
        ('NFTA_RULE_EXPRESSIONS', '*nft_expr'),
        ('NFTA_RULE_COMPAT', 'hex'),
        ('NFTA_RULE_POSITION', 'be64'),
        ('NFTA_RULE_USERDATA', 'hex'),
        ('NFTA_RULE_PAD', 'hex'),
        ('NFTA_RULE_ID', 'be32'),
        ('NFTA_RULE_POSITION_ID', 'be32'),
        ('NFTA_RULE_CHAIN_ID', 'be32'),
    )


class nft_set_msg(nfgen_msg, nft_contains_expr):
    prefix = 'NFTA_SET_'
    nla_map = (
        ('NFTA_SET_UNSPEC', 'none'),
        ('NFTA_SET_TABLE', 'asciiz'),
        ('NFTA_SET_NAME', 'asciiz'),
        ('NFTA_SET_FLAGS', 'set_flags'),
        ('NFTA_SET_KEY_TYPE', 'be32'),
        ('NFTA_SET_KEY_LEN', 'be32'),
        ('NFTA_SET_DATA_TYPE', 'be32'),
        ('NFTA_SET_DATA_LEN', 'be32'),
        ('NFTA_SET_POLICY', 'set_policy'),
        ('NFTA_SET_DESC', 'set_desc'),
        ('NFTA_SET_ID', 'be32'),
        ('NFTA_SET_TIMEOUT', 'be64'),
        ('NFTA_SET_GC_INTERVAL', 'be32'),
        ('NFTA_SET_USERDATA', 'set_udata'),
        ('NFTA_SET_PAD', 'hex'),
        ('NFTA_SET_OBJ_TYPE', 'be32'),
        ('NFTA_SET_HANDLE', 'be64'),
        ('NFTA_SET_EXPR', 'nft_expr'),
        ('NFTA_SET_EXPRESSIONS', '*nft_expr'),
    )

    class set_udata(nftnl_udata):
        udata_types = (
            "NFTNL_UDATA_SET_KEYBYTEORDER",
            "NFTNL_UDATA_SET_DATABYTEORDER",
            "NFTNL_UDATA_SET_MERGE_ELEMENTS",
            "NFTNL_UDATA_SET_KEY_TYPEOF",
            "NFTNL_UDATA_SET_DATA_TYPEOF",
            "NFTNL_UDATA_SET_EXPR",
            "NFTNL_UDATA_SET_DATA_INTERVAL",
            "NFTNL_UDATA_SET_COMMENT",
        )

    class set_flags(nft_flags_be32):
        ops = (
            'NFT_SET_ANONYMOUS',
            'NFT_SET_CONSTANT',
            'NFT_SET_INTERVAL',
            'NFT_SET_MAP',
            'NFT_SET_TIMEOUT',
            'NFT_SET_EVAL',
            'NFT_SET_OBJECT',
            'NFT_SET_CONCAT',
        )

    class set_policy(nft_map_be32):
        ops = {0: 'NFT_SET_POL_PERFORMANCE', 1: 'NFT_SET_POL_MEMORY'}

    class set_desc(nla):
        nla_map = (
            ('NFTA_SET_DESC_UNSPEC', 'none'),
            ('NFTA_SET_DESC_SIZE', 'be32'),
            ('NFTA_SET_DESC_CONCAT', '*list_elem'),
        )

        class list_elem(nla):
            nla_map = (
                ('NFTA_LIST_UNSPEC', 'none'),
                ('NFTA_LIST_ELEM', '*set_field_attribute'),
            )

            class set_field_attribute(nla):
                nla_map = (
                    ('NFTA_SET_FIELD_UNSPEC', 'none'),
                    ('NFTA_SET_FIELD_LEN', 'be32'),
                )


class nft_table_msg(nfgen_msg, nft_contains_expr):
    prefix = 'NFTA_TABLE_'
    nla_map = (
        ('NFTA_TABLE_UNSPEC', 'none'),
        ('NFTA_TABLE_NAME', 'asciiz'),
        ('NFTA_TABLE_FLAGS', 'be32'),
        ('NFTA_TABLE_USE', 'be32'),
        ('NFTA_TABLE_HANDLE', 'be64'),
        ('NFTA_TABLE_PAD', 'hex'),
        ('NFTA_TABLE_USERDATA', 'hex'),
    )


class nft_set_elem_list_msg(nfgen_msg):
    prefix = 'NFTA_SET_ELEM_LIST_'
    nla_map = (
        ('NFTA_SET_ELEM_LIST_UNSPEC', 'none'),
        ('NFTA_SET_ELEM_LIST_TABLE', 'asciiz'),
        ('NFTA_SET_ELEM_LIST_SET', 'asciiz'),
        ('NFTA_SET_ELEM_LIST_ELEMENTS', '*set_elem'),
        ('NFTA_SET_ELEM_LIST_SET_ID', 'be32'),
    )

    class set_elem(nla, nft_contains_expr):
        nla_map = (
            ('NFTA_SET_ELEM_UNSPEC', 'none'),
            ('NFTA_SET_ELEM_KEY', 'data_attributes'),
            ('NFTA_SET_ELEM_DATA', 'data_attributes'),
            ('NFTA_SET_ELEM_FLAGS', 'set_elem_flags'),
            ('NFTA_SET_ELEM_TIMEOUT', 'be64'),
            ('NFTA_SET_ELEM_EXPIRATION', 'be64'),
            ('NFTA_SET_ELEM_USERDATA', 'set_elem_udata'),
            ('NFTA_SET_ELEM_EXPR', 'nft_expr'),
            ('NFTA_SET_ELEM_PAD', 'hex'),
            ('NFTA_SET_ELEM_OBJREF', 'asciiz'),
            ('NFTA_SET_ELEM_KEY_END', 'data_attributes'),
            ('NFTA_SET_ELEM_EXPRESSIONS', '*nft_expr'),
        )

        class set_elem_udata(nftnl_udata):
            udata_types = (
                "NFTNL_UDATA_SET_ELEM_COMMENT",
                "NFTNL_UDATA_SET_ELEM_FLAGS",
            )

        class set_elem_flags(nft_flags_be32):
            ops = {1: 'NFT_SET_ELEM_INTERVAL_END'}

        class data_attributes(nla):
            nla_map = (
                ('NFTA_DATA_UNSPEC', 'none'),
                ('NFTA_DATA_VALUE', 'binary'),
                ('NFTA_DATA_VERDICT', 'verdict_attributes'),
            )

            class verdict_attributes(nla):
                nla_map = (
                    ('NFTA_VERDICT_UNSPEC', 'none'),
                    ('NFTA_VERDICT_CODE', 'verdict_code'),
                    ('NFTA_VERDICT_CHAIN', 'asciiz'),
                    ('NFTA_VERDICT_CHAIN_ID', 'be32'),
                )

                class verdict_code(nft_map_be32_signed):
                    ops = {
                        0: 'NF_DROP',
                        1: 'NF_ACCEPT',
                        2: 'NF_STOLEN',
                        3: 'NF_QUEUE',
                        4: 'NF_REPEAT',
                        5: 'NF_STOP',
                        -1: 'NFT_CONTINUE',
                        -2: 'NFT_BREAK',
                        -3: 'NFT_JUMP',
                        -4: 'NFT_GOTO',
                        -5: 'NFT_RETURN',
                    }


class nft_flowtable_msg(nfgen_msg):
    prefix = 'NFTA_FLOWTABLE_'
    nla_map = (
        ('NFTA_FLOWTABLE_UNSPEC', 'none'),
        ('NFTA_FLOWTABLE_TABLE', 'asciiz'),
        ('NFTA_FLOWTABLE_NAME', 'asciiz'),
        ('NFTA_FLOWTABLE_HOOK', 'flowtable_hook'),
        ('NFTA_FLOWTABLE_USE', 'be32'),
        ('NFTA_FLOWTABLE_HANDLE', 'be64'),
        ('NFTA_FLOWTABLE_PAD', 'hex'),
        ('NFTA_FLOWTABLE_FLAGS', 'nft_flowtable_flags'),
    )

    class nft_flowtable_flags(nft_flags_be32):
        ops = ('NFT_FLOWTABLE_HW_OFFLOAD', 'NFT_FLOWTABLE_COUNTER')

    class flowtable_hook(nft_device):
        nla_map = (
            ('NFTA_FLOWTABLE_HOOK_UNSPEC', 'none'),
            ('NFTA_FLOWTABLE_HOOK_NUM', 'be32'),
            ('NFTA_FLOWTABLE_HOOK_PRIORITY', 'be32'),
            ('NFTA_FLOWTABLE_HOOK_DEVS', 'device_attributes'),
        )


class NFTSocket(NetlinkSocket):
    '''
    NFNetlink socket (family=NETLINK_NETFILTER).

    Implements API to the nftables functionality.
    '''

    policy = {
        NFT_MSG_NEWTABLE: nft_table_msg,
        NFT_MSG_GETTABLE: nft_table_msg,
        NFT_MSG_DELTABLE: nft_table_msg,
        NFT_MSG_NEWCHAIN: nft_chain_msg,
        NFT_MSG_GETCHAIN: nft_chain_msg,
        NFT_MSG_DELCHAIN: nft_chain_msg,
        NFT_MSG_NEWRULE: nft_rule_msg,
        NFT_MSG_GETRULE: nft_rule_msg,
        NFT_MSG_DELRULE: nft_rule_msg,
        NFT_MSG_NEWSET: nft_set_msg,
        NFT_MSG_GETSET: nft_set_msg,
        NFT_MSG_DELSET: nft_set_msg,
        NFT_MSG_NEWGEN: nft_gen_msg,
        NFT_MSG_GETGEN: nft_gen_msg,
        NFT_MSG_NEWSETELEM: nft_set_elem_list_msg,
        NFT_MSG_GETSETELEM: nft_set_elem_list_msg,
        NFT_MSG_DELSETELEM: nft_set_elem_list_msg,
        NFT_MSG_NEWFLOWTABLE: nft_flowtable_msg,
        NFT_MSG_GETFLOWTABLE: nft_flowtable_msg,
        NFT_MSG_DELFLOWTABLE: nft_flowtable_msg,
    }

    def __init__(self, version=1, attr_revision=0, nfgen_family=2):
        super(NFTSocket, self).__init__(family=NETLINK_NETFILTER)
        policy = dict(
            [
                (x | (NFNL_SUBSYS_NFTABLES << 8), y)
                for (x, y) in self.policy.items()
            ]
        )
        self.register_policy(policy)
        self._proto_version = version
        self._attr_revision = attr_revision
        self._nfgen_family = nfgen_family
        self._ts = threading.local()
        self._write_lock = threading.RLock()

    def begin(self):
        with self._write_lock:
            if hasattr(self._ts, 'data'):
                # transaction is already started
                return False

            self._ts.data = b''
            self._ts.seqnum = (
                self.addr_pool.alloc(),  # begin
                self.addr_pool.alloc(),  # tx
                self.addr_pool.alloc(),
            )  # commit
            msg = nfgen_msg()
            msg['res_id'] = NFNL_SUBSYS_NFTABLES
            msg['header']['type'] = 0x10
            msg['header']['flags'] = NLM_F_REQUEST
            msg['header']['sequence_number'] = self._ts.seqnum[0]
            msg.encode()
            self._ts.data += msg.data
            return True

    def commit(self):
        with self._write_lock:
            msg = nfgen_msg()
            msg['res_id'] = NFNL_SUBSYS_NFTABLES
            msg['header']['type'] = 0x11
            msg['header']['flags'] = NLM_F_REQUEST
            msg['header']['sequence_number'] = self._ts.seqnum[2]
            msg.encode()
            self._ts.data += msg.data
            self.sendto(self._ts.data, (0, 0))
            for seqnum in self._ts.seqnum:
                self.addr_pool.free(seqnum, ban=10)
            del self._ts.data

    def request_get(
        self,
        msg,
        msg_type,
        msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
        terminate=None,
    ):
        '''
        Read-only requests do not require transactions. Just run
        the request and get an answer.
        '''
        msg['nfgen_family'] = self._nfgen_family
        return tuple(
            self.nlm_request(
                msg,
                msg_type | (NFNL_SUBSYS_NFTABLES << 8),
                msg_flags,
                terminate=terminate,
            )
        )

    def request_put(self, msg, msg_type, msg_flags=NLM_F_REQUEST):
        '''
        Read-write requests.
        '''
        one_shot = self.begin()
        msg['header']['type'] = (NFNL_SUBSYS_NFTABLES << 8) | msg_type
        msg['header']['flags'] = msg_flags
        msg['header']['sequence_number'] = self._ts.seqnum[1]
        msg['nfgen_family'] = self._nfgen_family
        msg.encode()
        self._ts.data += msg.data
        if one_shot:
            self.commit()

    def _command(self, msg_class, commands, cmd, kwarg):
        flags = kwarg.pop('flags', NLM_F_ACK)
        cmd_name = cmd
        cmd_flags = {
            'add': NLM_F_CREATE | NLM_F_APPEND,
            'create': NLM_F_CREATE | NLM_F_APPEND | NLM_F_EXCL,
            'insert': NLM_F_CREATE,
            'replace': NLM_F_REPLACE,
        }
        flags |= cmd_flags.get(cmd, 0)
        flags |= NLM_F_REQUEST
        cmd = commands[cmd]
        msg = msg_class()
        msg['attrs'] = []
        #
        # a trick to pass keyword arguments as On rderedDict instance:
        #
        # ordered_args = OrderedDict()
        # ordered_args['arg1'] = value1
        # ordered_args['arg2'] = value2
        # ...
        # nft.rule('add', kwarg=ordered_args)
        #
        if 'kwarg' in kwarg:
            kwarg = kwarg['kwarg']
        #
        for key, value in kwarg.items():
            nla = msg_class.name2nla(key)
            msg['attrs'].append([nla, value])
        msg['header']['type'] = (NFNL_SUBSYS_NFTABLES << 8) | cmd
        msg['header']['flags'] = flags | NLM_F_REQUEST
        msg['nfgen_family'] = self._nfgen_family

        if cmd_name != 'get':
            trans_start = nfgen_msg()
            trans_start['res_id'] = NFNL_SUBSYS_NFTABLES
            trans_start['header']['type'] = 0x10
            trans_start['header']['flags'] = NLM_F_REQUEST

            trans_end = nfgen_msg()
            trans_end['res_id'] = NFNL_SUBSYS_NFTABLES
            trans_end['header']['type'] = 0x11
            trans_end['header']['flags'] = NLM_F_REQUEST

            messages = [trans_start, msg, trans_end]
            self.nlm_request_batch(messages, noraise=(flags & NLM_F_ACK) == 0)
            # Only throw an error when the request fails. For now,
            # do not return anything.
        else:
            return self.request_get(msg, msg['header']['type'], flags)[0]


# call nft describe "data_type" for more informations
DATA_TYPE_NAME_TO_INFO = {
    "verdict": (DATA_TYPE_VERDICT, 4, nft_data.nfta_data.verdict.verdict_code),
    "nf_proto": (DATA_TYPE_NFPROTO, 1, nlmsg_atoms.uint8),
    "bitmask": (DATA_TYPE_BITMASK, 4, nlmsg_atoms.uint32),
    "integer": (DATA_TYPE_INTEGER, 4, nlmsg_atoms.int32),
    "string": (DATA_TYPE_STRING, 0, nlmsg_atoms.asciiz),
    "lladdr": (DATA_TYPE_LLADDR, 0, nlmsg_atoms.lladdr),
    "ipv4_addr": (DATA_TYPE_IPADDR, 4, nlmsg_atoms.ip4addr),
    "ipv6_addr": (DATA_TYPE_IP6ADDR, 16, nlmsg_atoms.ip6addr),
    "ether_addr": (DATA_TYPE_ETHERADDR, 6, nlmsg_atoms.l2addr),
    "ether_type": (DATA_TYPE_ETHERADDR, 2, nlmsg_atoms.uint16),
    "inet_proto": (DATA_TYPE_INET_PROTOCOL, 1, nlmsg_atoms.uint8),
}
DATA_TYPE_ID_TO_NAME = {
    value[0]: key for key, value in DATA_TYPE_NAME_TO_INFO.items()
}



'''
u32
+++

Filters can take an `action` argument, which affects the packet
behavior when the filter matches. Currently the gact, bpf, and police
action types are supported, and can be attached to the u32 and bpf
filter types::

    # An action can be a simple string, which translates to a gact type
    action = "drop"

    # Or it can be an explicit type (these are equivalent)
    action = dict(kind="gact", action="drop")

    # There can also be a chain of actions, which depend on the return
    # value of the previous action.
    action = [
        dict(kind="bpf", fd=fd, name=name, action="ok"),
        dict(kind="police", rate="10kbit", burst=10240, limit=0),
        dict(kind="gact", action="ok"),
    ]

    # Add the action to a u32 match-all filter
    ip.tc("add", "htb", eth0, 0x10000, default=0x200000)
    ip.tc("add-filter", "u32", eth0,
          parent=0x10000,
          prio=10,
          protocol=protocols.ETH_P_ALL,
          target=0x10020,
          keys=["0x0/0x0+0"],
          action=action)

    # Add two more filters: One to send packets with a src address of
    # 192.168.0.1/32 into 1:10 and the second to send packets  with a
    # dst address of 192.168.0.0/24 into 1:20
    ip.tc("add-filter", "u32", eth0,
        parent=0x10000,
        prio=10,
        protocol=protocols.ETH_P_IP,
        target=0x10010,
        keys=["0xc0a80001/0xffffffff+12"])
        # 0xc0a800010 = 192.168.0.1
        # 0xffffffff = 255.255.255.255 (/32)
        # 12 = Source network field bit offset

    ip.tc("add-filter", "u32", eth0,
        parent=0x10000,
        prio=10,
        protocol=protocols.ETH_P_IP,
        target=0x10020,
        keys=["0xc0a80000/0xffffff00+16"])
        # 0xc0a80000 = 192.168.0.0
        # 0xffffff00 = 255.255.255.0 (/24)
        # 16 = Destination network field bit offset
'''
import struct
from socket import htons

from pyroute2.netlink import nla, nlmsg
from pyroute2.netlink.rtnl.tcmsg.act_police import (
    get_parameters as ap_parameters,
)
from pyroute2.netlink.rtnl.tcmsg.act_police import nla_plus_police
from pyroute2.netlink.rtnl.tcmsg.common_act import get_tca_action, tca_act_prio


def fix_msg(msg, kwarg):
    msg['info'] = htons(kwarg.get('protocol', 0) & 0xFFFF) | (
        (kwarg.get('prio', 0) << 16) & 0xFFFF0000
    )


def get_parameters(kwarg):
    ret = {'attrs': []}

    if kwarg.get('rate'):
        ret['attrs'].append(['TCA_U32_POLICE', ap_parameters(kwarg)])
    elif kwarg.get('action'):
        ret['attrs'].append(['TCA_U32_ACT', get_tca_action(kwarg)])

    ret['attrs'].append(['TCA_U32_CLASSID', kwarg['target']])
    ret['attrs'].append(['TCA_U32_SEL', {'keys': kwarg['keys']}])

    return ret


class options(nla, nla_plus_police):
    nla_map = (
        ('TCA_U32_UNSPEC', 'none'),
        ('TCA_U32_CLASSID', 'uint32'),
        ('TCA_U32_HASH', 'uint32'),
        ('TCA_U32_LINK', 'hex'),
        ('TCA_U32_DIVISOR', 'uint32'),
        ('TCA_U32_SEL', 'u32_sel'),
        ('TCA_U32_POLICE', 'police'),
        ('TCA_U32_ACT', 'tca_act_prio'),
        ('TCA_U32_INDEV', 'hex'),
        ('TCA_U32_PCNT', 'u32_pcnt'),
        ('TCA_U32_MARK', 'u32_mark'),
    )

    tca_act_prio = tca_act_prio

    class u32_sel(nla):
        fields = (
            ('flags', 'B'),
            ('offshift', 'B'),
            ('nkeys', 'B'),
            ('__align', 'x'),
            ('offmask', '>H'),
            ('off', 'H'),
            ('offoff', 'h'),
            ('hoff', 'h'),
            ('hmask', '>I'),
        )

        class u32_key(nlmsg):
            header = None
            fields = (
                ('key_mask', '>I'),
                ('key_val', '>I'),
                ('key_off', 'i'),
                ('key_offmask', 'i'),
            )

        def encode(self):
            '''
            Key sample::

                'keys': ['0x0006/0x00ff+8',
                         '0x0000/0xffc0+2',
                         '0x5/0xf+0',
                         '0x10/0xff+33']

                => 00060000/00ff0000 + 8
                   05000000/0f00ffc0 + 0
                   00100000/00ff0000 + 32
            '''

            def cut_field(key, separator):
                '''
                split a field from the end of the string
                '''
                field = '0'
                pos = key.find(separator)
                new_key = key
                if pos > 0:
                    field = key[pos + 1 :]
                    new_key = key[:pos]
                return (new_key, field)

            # 'header' array to pack keys to
            header = [(0, 0) for i in range(256)]

            keys = []
            # iterate keys and pack them to the 'header'
            for key in self['keys']:
                # TODO tags: filter
                (key, nh) = cut_field(key, '@')  # FIXME: do not ignore nh
                (key, offset) = cut_field(key, '+')
                offset = int(offset, 0)
                # a little trick: if you provide /00ff+8, that
                # really means /ff+9, so we should take it into
                # account
                (key, mask) = cut_field(key, '/')
                if mask[:2] == '0x':
                    mask = mask[2:]
                    while True:
                        if mask[:2] == '00':
                            offset += 1
                            mask = mask[2:]
                        else:
                            break
                    mask = '0x' + mask
                mask = int(mask, 0)
                value = int(key, 0)
                bits = 24
                if mask == 0 and value == 0:
                    key = self.u32_key(data=self.data)
                    key['key_off'] = offset
                    key['key_mask'] = mask
                    key['key_val'] = value
                    keys.append(key)
                for bmask in struct.unpack('4B', struct.pack('>I', mask)):
                    if bmask > 0:
                        bvalue = (value & (bmask << bits)) >> bits
                        header[offset] = (bvalue, bmask)
                        offset += 1
                    bits -= 8

            # recalculate keys from 'header'
            key = None
            value = 0
            mask = 0
            for offset in range(256):
                (bvalue, bmask) = header[offset]
                if bmask > 0 and key is None:
                    key = self.u32_key(data=self.data)
                    key['key_off'] = offset
                    key['key_mask'] = 0
                    key['key_val'] = 0
                    bits = 24
                if key is not None and bits >= 0:
                    key['key_mask'] |= bmask << bits
                    key['key_val'] |= bvalue << bits
                    bits -= 8
                    if bits < 0 or offset == 255:
                        keys.append(key)
                        key = None

            if not keys:
                raise ValueError('no keys specified')
            self['nkeys'] = len(keys)
            # FIXME: do not hardcode flags :)
            self['flags'] = 1

            nla.encode(self)
            offset = self.offset + 20  # 4 bytes header + 16 bytes fields
            for key in keys:
                key.offset = offset
                key.encode()
                offset += 16  # keys haven't header
            self.length = offset - self.offset
            struct.pack_into('H', self.data, self.offset, offset - self.offset)

        def decode(self):
            nla.decode(self)
            offset = self.offset + 16
            self['keys'] = []
            nkeys = self['nkeys']
            while nkeys:
                key = self.u32_key(data=self.data, offset=offset)
                key.decode()
                offset += 16
                self['keys'].append(key)
                nkeys -= 1

    class u32_mark(nla):
        fields = (('val', 'I'), ('mask', 'I'), ('success', 'I'))

    class u32_pcnt(nla):
        fields = (('rcnt', 'Q'), ('rhit', 'Q'), ('kcnts', 'Q'))



import json
import select
import sys

from pyroute2.dhcp import (
    BOOTREQUEST,
    DHCPACK,
    DHCPDISCOVER,
    DHCPOFFER,
    DHCPREQUEST,
)
from pyroute2.dhcp.dhcp4msg import dhcp4msg
from pyroute2.dhcp.dhcp4socket import DHCP4Socket


def req(s, poll, msg, expect):
    do_req = True
    xid = None

    while True:
        # get transaction id
        if do_req:
            xid = s.put(msg)['xid']
        # wait for response
        events = poll.poll(2)
        for fd, event in events:
            response = s.get()
            if response['xid'] != xid:
                do_req = False
                continue
            if response['options']['message_type'] != expect:
                raise Exception("DHCP protocol error")
            return response
        do_req = True


def action(ifname):
    s = DHCP4Socket(ifname)
    poll = select.poll()
    poll.register(s, select.POLLIN | select.POLLPRI)

    # DISCOVER
    discover = dhcp4msg(
        {
            'op': BOOTREQUEST,
            'chaddr': s.l2addr,
            'options': {
                'message_type': DHCPDISCOVER,
                'parameter_list': [1, 3, 6, 12, 15, 28],
            },
        }
    )
    reply = req(s, poll, discover, expect=DHCPOFFER)

    # REQUEST
    request = dhcp4msg(
        {
            'op': BOOTREQUEST,
            'chaddr': s.l2addr,
            'options': {
                'message_type': DHCPREQUEST,
                'requested_ip': reply['yiaddr'],
                'server_id': reply['options']['server_id'],
                'parameter_list': [1, 3, 6, 12, 15, 28],
            },
        }
    )
    reply = req(s, poll, request, expect=DHCPACK)
    s.close()
    return reply


def run():
    if len(sys.argv) > 1:
        ifname = sys.argv[1]
    else:
        ifname = 'eth0'
    print(json.dumps(action(ifname), indent=4))


if __name__ == '__main__':
    run()



'''
.. note:: New in verision 0.5.11

.. testsetup::

    from pyroute2 import NDB
    ndb = NDB(sources=[{'target': 'localhost', 'kind': 'IPMock'}])

.. testcleanup:: *

    for key, value in tuple(globals().items()):
        if key.startswith('ndb') and hasattr(value, 'close'):
            value.close()

Filtering example:

.. testcode::

    report = ndb.interfaces.dump()
    report.select_fields('index', 'ifname', 'address', 'state')
    report.transform_fields(
        address=lambda r: '%s%s.%s%s.%s%s' % tuple(r.address.split(':'))
    )
    for record in report.format('csv'):
        print(record)

.. testoutput::

    'index','ifname','address','state'
    1,'lo','0000.0000.0000','up'
    2,'eth0','5254.0072.58b2','up'

'''
import json
import warnings
from itertools import chain

from pyroute2 import cli

MAX_REPORT_LINES = 10000

deprecation_notice = '''
RecordSet API is deprecated, pls refer to:

https://docs.pyroute2.org/ndb_reports.html
'''


def format_json(dump, headless=False):
    buf = []
    fnames = None
    yield '['
    for record in dump:
        if fnames is None:
            if headless:
                fnames = record._names
            else:
                fnames = record
                continue
        if buf:
            buf[-1] += ','
            for line in buf:
                yield line
            buf = []
        lines = json.dumps(dict(zip(fnames, record)), indent=4).split('\n')
        buf.append('    {')
        for line in sorted(lines[1:-1]):
            if line[-1] == ',':
                line = line[:-1]
            buf.append('    %s,' % line)
        buf[-1] = buf[-1][:-1]
        buf.append('    }')
    for line in buf:
        yield line
    yield ']'


def format_csv(dump, headless=False):
    def dump_record(rec):
        row = []
        for field in rec:
            if isinstance(field, int):
                row.append('%i' % field)
            elif field is None:
                row.append('')
            else:
                row.append("'%s'" % field)
        return row

    fnames = None
    for record in dump:
        if fnames is None and headless:
            fnames = True
            yield ','.join(dump_record(record._names))
        yield ','.join(dump_record(record))


class Record:
    def __init__(self, names, values, ref_class=None):
        self._names = tuple(names)
        self._values = tuple(values)
        if len(self._names) != len(self._values):
            raise ValueError('names and values must have the same length')
        self._ref_class = ref_class

    def __getitem__(self, key):
        idx = len(self._names)
        for i in reversed(self._names):
            idx -= 1
            if i == key:
                return self._values[idx]

    def __setitem__(self, *argv, **kwarg):
        raise TypeError('immutable object')

    def __getattribute__(self, key):
        if key.startswith('_'):
            return object.__getattribute__(self, key)
        else:
            return self[key]

    def __setattr__(self, key, value):
        if not key.startswith('_'):
            raise TypeError('immutable object')
        return object.__setattr__(self, key, value)

    def __iter__(self):
        return iter(self._values)

    def __repr__(self):
        return repr(self._values)

    def __len__(self):
        return len(self._values)

    def _select_fields(self, *fields):
        return Record(fields, map(lambda x: self[x], fields), self._ref_class)

    def _transform_fields(self, **spec):
        data = self._as_dict()
        for key, func in spec.items():
            data[key] = func(self)
        return Record(data.keys(), data.values(), self._ref_class)

    def _match(self, f=None, **spec):
        if callable(f):
            return f(self)
        for key, value in spec.items():
            if not (
                value(self[key]) if callable(value) else (self[key] == value)
            ):
                return False
        return True

    def _as_dict(self):
        ret = {}
        for key, value in zip(self._names, self._values):
            ret[key] = value
        return ret

    def __eq__(self, right):
        if hasattr(right, '_names'):
            n = all(x[0] == x[1] for x in zip(self._names, right._names))
            v = all(x[0] == x[1] for x in zip(self._values, right._values))
            return n and v
        elif isinstance(right, dict):
            for key, value in right.items():
                if value != self[key]:
                    break
            else:
                return True
            return False
        elif self._ref_class is not None and isinstance(right, (str, int)):
            return self._ref_class.compare_record(self, right)
        else:
            return all(x[0] == x[1] for x in zip(self._values, right))


class BaseRecordSet(object):
    def __init__(self, generator, ellipsis='(...)'):
        self.generator = generator
        self.ellipsis = ellipsis

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.generator)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __repr__(self):
        counter = 0
        ret = []
        for record in self:
            if isinstance(record, str):
                ret.append(record)
            else:
                ret.append(repr(record))
            ret.append('\n')
            counter += 1
            if self.ellipsis and counter > MAX_REPORT_LINES:
                ret.append(self.ellipsis)
                break
        if ret:
            ret.pop()
        return ''.join(ret)


class RecordSet(BaseRecordSet):
    '''
    NDB views return objects of this class with `summary()` and `dump()`
    methods. RecordSet objects are generator-based, they do not store the
    data in the memory, but transform them on the fly.

    RecordSet filters also return objects of this class, thus making possible
    to make chains of filters.
    '''

    def __init__(self, generator, ellipsis=True):
        super().__init__(generator, ellipsis)
        self.filters = []

    def __next__(self):
        while True:
            record = next(self.generator)
            for f in self.filters:
                record = f(record)
                if record is None:
                    break
            else:
                return record

    def select_fields(self, *fields):
        '''
        Select only chosen fields for every record:

        .. testcode::

            report = ndb.interfaces.dump()
            report.select_fields('index', 'ifname')
            for line in report.format('csv'):
                print(line)

        .. testoutput::

            'index','ifname'
            1,'lo'
            2,'eth0'
        '''
        self.filters.append(lambda x: x._select_fields(*fields))

    def select_records(self, f=None, **spec):
        '''
        Select records based on a function f() or a spec match. A spec
        is dictionary of pairs `field: constant` or `field: callable`:

        .. testcode::

            report = ndb.addresses.summary()
            report.select_records(ifname=lambda x: x.startswith('eth'))
            for line in report.format('csv'):
                print(line)

        .. testoutput::

            'target','tflags','ifname','address','prefixlen'
            'localhost',0,'eth0','192.168.122.28',24
        '''
        self.filters.append(lambda x: x if x._match(f, **spec) else None)

    def transform_fields(self, **kwarg):
        '''
        Transform fields with a function. Function must accept
        the record as the only argument:

        .. testcode::

            report = ndb.addresses.summary()
            report.transform_fields(
                address=lambda r: f'{r.address}/{r.prefixlen}'
            )
            report.select_fields('ifname', 'address')
            for line in report.format('csv'):
                print(line)

        .. testoutput::

            'ifname','address'
            'lo','127.0.0.1/8'
            'eth0','192.168.122.28/24'
        '''
        self.filters.append(lambda x: x._transform_fields(**kwarg))

    @cli.show_result
    def transform(self, **kwarg):
        warnings.warn(deprecation_notice, DeprecationWarning)

        def g():
            for record in self.generator:
                if isinstance(record, Record):
                    values = []
                    names = record._names
                    for name, value in zip(names, record._values):
                        if name in kwarg:
                            value = kwarg[name](value)
                        values.append(value)
                    record = Record(names, values, record._ref_class)
                yield record

        return RecordSet(g())

    @cli.show_result
    def filter(self, f=None, **kwarg):
        warnings.warn(deprecation_notice, DeprecationWarning)

        def g():
            for record in self.generator:
                m = True
                for key in kwarg:
                    if kwarg[key] != getattr(record, key):
                        m = False
                if m:
                    if f is None:
                        yield record
                    elif f(record):
                        yield record

        return RecordSet(g())

    @cli.show_result
    def select(self, *argv):
        warnings.warn(deprecation_notice, DeprecationWarning)
        return self.fields(*argv)

    @cli.show_result
    def fields(self, *fields):
        warnings.warn(deprecation_notice, DeprecationWarning)

        def g():
            for record in self.generator:
                yield record._select_fields(*fields)

        return RecordSet(g())

    @cli.show_result
    def join(self, right, condition=lambda r1, r2: True, prefix=''):
        warnings.warn(deprecation_notice, DeprecationWarning)
        # fetch all the records from the right
        # ACHTUNG it may consume a lot of memory
        right = tuple(right)

        def g():
            for r1 in self.generator:
                for r2 in right:
                    if condition(r1, r2):
                        n = tuple(
                            chain(
                                r1._names,
                                ['%s%s' % (prefix, x) for x in r2._names],
                            )
                        )
                        v = tuple(chain(r1._values, r2._values))
                        yield Record(n, v, r1._ref_class)

        return RecordSet(g())

    @cli.show_result
    def format(self, kind):
        '''
        Return an iterator over text lines in the chosen format.

        Supported formats: 'json', 'csv'.
        '''
        if kind == 'json':
            return BaseRecordSet(format_json(self, headless=True))
        elif kind == 'csv':
            return BaseRecordSet(format_csv(self, headless=True))
        else:
            raise ValueError()

    def count(self):
        '''
        Return number of records.

        The method exhausts the generator.
        '''
        counter = 0
        for record in self:
            counter += 1
        return counter

    def __getitem__(self, key):
        return list(self)[key]



'''
NetNS objects
=============

A NetNS object is IPRoute-like. It runs in the main network
namespace, but also creates a proxy process running in
the required netns. All the netlink requests are done via
that proxy process.

NetNS supports standard IPRoute API, so can be used instead
of IPRoute, e.g., in IPDB::

    # start the main network settings database:
    ipdb_main = IPDB()
    # start the same for a netns:
    ipdb_test = IPDB(nl=NetNS('test'))

    # create VETH
    ipdb_main.create(ifname='v0p0', kind='veth', peer='v0p1').commit()

    # move peer VETH into the netns
    with ipdb_main.interfaces.v0p1 as veth:
        veth.net_ns_fd = 'test'

    # please keep in mind, that netns move clears all the settings
    # on a VETH interface pair, so one should run netns assignment
    # as a separate operation only

    # assign addresses
    # please notice, that `v0p1` is already in the `test` netns,
    # so should be accessed via `ipdb_test`
    with ipdb_main.interfaces.v0p0 as veth:
        veth.add_ip('172.16.200.1/24')
        veth.up()
    with ipdb_test.interfaces.v0p1 as veth:
        veth.add_ip('172.16.200.2/24')
        veth.up()

Please review also the test code, under `tests/test_netns.py` for
more examples.

By default, NetNS creates requested netns, if it doesn't exist,
or uses existing one. To control this behaviour, one can use flags
as for `open(2)` system call::

    # create a new netns or fail, if it already exists
    netns = NetNS('test', flags=os.O_CREAT | os.O_EXCL)

    # create a new netns or use existing one
    netns = NetNS('test', flags=os.O_CREAT)

    # the same as above, the default behaviour
    netns = NetNS('test')

To remove a network namespace::

    from pyroute2 import NetNS
    netns = NetNS('test')
    netns.close()
    netns.remove()

One should stop it first with `close()`, and only after that
run `remove()`.

'''

import atexit
import errno
import logging
import os
from functools import partial

from pyroute2.iproute import RTNL_API
from pyroute2.netlink.rtnl import RTMGRP_DEFAULTS
from pyroute2.netlink.rtnl.iprsocket import MarshalRtnl
from pyroute2.netns import remove, setns

from ..remote.transport import RemoteSocket, Server, Transport

log = logging.getLogger(__name__)


class FD(object):
    def __init__(self, fd):
        self.fd = fd
        for name in ('read', 'write', 'close'):
            setattr(self, name, partial(getattr(os, name), self.fd))

    def fileno(self):
        return self.fd

    def flush(self):
        return None


class NetNS(RTNL_API, RemoteSocket):
    '''
    NetNS is the IPRoute API with network namespace support.

    **Why not IPRoute?**

    The task to run netlink commands in some network namespace, being in
    another network namespace, requires the architecture, that differs
    too much from a simple Netlink socket.

    NetNS starts a proxy process in a network namespace and uses
    `multiprocessing` communication channels between the main and the proxy
    processes to route all `recv()` and `sendto()` requests/responses.

    **Any specific API calls?**

    Nope. `NetNS` supports all the same, that `IPRoute` does, in the same
    way. It provides full `socket`-compatible API and can be used in
    poll/select as well.

    The only difference is the `close()` call. In the case of `NetNS` it
    is **mandatory** to close the socket before exit.

    '''

    def __init__(
        self,
        netns,
        flags=os.O_CREAT,
        target=None,
        libc=None,
        groups=RTMGRP_DEFAULTS,
    ):
        self.netns = netns
        self.flags = flags
        target = target or netns
        trnsp_in, self.remote_trnsp_out = [Transport(FD(x)) for x in os.pipe()]
        self.remote_trnsp_in, trnsp_out = [Transport(FD(x)) for x in os.pipe()]

        self.child = os.fork()
        if self.child == 0:
            # child process
            trnsp_in.close()
            trnsp_out.close()
            trnsp_in.file_obj.close()
            trnsp_out.file_obj.close()
            try:
                setns(self.netns, self.flags, libc=libc)
            except OSError as e:
                (self.remote_trnsp_out.send({'stage': 'init', 'error': e}))
                os._exit(e.errno)
            except Exception as e:
                (
                    self.remote_trnsp_out.send(
                        {
                            'stage': 'init',
                            'error': OSError(errno.ECOMM, str(e), self.netns),
                        }
                    )
                )
                os._exit(255)

            try:
                Server(
                    self.remote_trnsp_in,
                    self.remote_trnsp_out,
                    target=target,
                    groups=groups,
                )
            finally:
                os._exit(0)

        try:
            self.remote_trnsp_in.close()
            self.remote_trnsp_out.close()
            super(NetNS, self).__init__(trnsp_in, trnsp_out, groups=groups)
            self.target = target
        except Exception:
            self.close()
            raise
        atexit.register(self.close)
        self.marshal = MarshalRtnl()

    def clone(self):
        return type(self)(self.netns, self.flags)

    def _cleanup_atexit(self):
        if hasattr(atexit, 'unregister'):
            atexit.unregister(self.close)
        else:
            try:
                atexit._exithandlers.remove((self.close, (), {}))
            except ValueError:
                pass

    def close(self, code=errno.ECONNRESET):
        self._cleanup_atexit()
        try:
            super(NetNS, self).close(code=code)
        except:
            # something went wrong, force server shutdown
            try:
                self.trnsp_out.send({'stage': 'shutdown'})
            except Exception:
                pass
            log.error('forced shutdown procedure, clean up netns manually')

    def open_file(self, path):
        '''Proxy the open_file method if we are the parent.'''
        if self.child != 0:
            return self.proxy('open_file', path)

        return super(NetNS, self).open_file(path)

    def close_file(self, fd):
        '''Proxy the close_file method if we are the parent.'''
        if self.child != 0:
            return self.proxy('close_file', fd)

        return super(NetNS, self).close_file(fd)

    def get_pid(self):
        '''Proxy the get_pid method if we are the parent.'''
        if self.child != 0:
            return self.proxy('get_pid')

        return super(NetNS, self).get_pid()

    def post_init(self):
        pass

    def remove(self):
        '''
        Try to remove this network namespace from the system.
        '''
        remove(self.netns)



try:
    from .iproute import RemoteIPRoute
except ImportError:
    from pyroute2.common import failed_class

    RemoteIPRoute = failed_class('mitogen library is not installed')

classes = [RemoteIPRoute]



from socket import AF_INET, inet_ntop, inet_pton

from pyroute2.dhcp import dhcpmsg, option


class dhcp4msg(dhcpmsg):
    #
    # https://www.ietf.org/rfc/rfc2131.txt
    #
    fields = (
        ('op', 'uint8', 1),  # request
        ('htype', 'uint8', 1),  # ethernet
        ('hlen', 'uint8', 6),  # ethernet addr len
        ('hops', 'uint8'),
        ('xid', 'uint32'),
        ('secs', 'uint16'),
        ('flags', 'uint16'),
        ('ciaddr', 'ip4addr'),
        ('yiaddr', 'ip4addr'),
        ('siaddr', 'ip4addr'),
        ('giaddr', 'ip4addr'),
        ('chaddr', 'l2paddr'),
        ('sname', '64s'),
        ('file', '128s'),
        ('cookie', '4s', b'c\x82Sc'),
    )
    #
    # https://www.ietf.org/rfc/rfc2132.txt
    #
    options = (
        (0, 'pad', 'none'),
        (1, 'subnet_mask', 'ip4addr'),
        (2, 'time_offset', 'be32'),
        (3, 'router', 'ip4list'),
        (4, 'time_server', 'ip4list'),
        (5, 'ien_name_server', 'ip4list'),
        (6, 'name_server', 'ip4list'),
        (7, 'log_server', 'ip4list'),
        (8, 'cookie_server', 'ip4list'),
        (9, 'lpr_server', 'ip4list'),
        (50, 'requested_ip', 'ip4addr'),
        (51, 'lease_time', 'be32'),
        (53, 'message_type', 'uint8'),
        (54, 'server_id', 'ip4addr'),
        (55, 'parameter_list', 'array8'),
        (57, 'messagi_size', 'be16'),
        (58, 'renewal_time', 'be32'),
        (59, 'rebinding_time', 'be32'),
        (60, 'vendor_id', 'string'),
        (61, 'client_id', 'client_id'),
        (255, 'end', 'none'),
    )

    class ip4addr(option):
        policy = {
            'format': '4s',
            'encode': lambda x: inet_pton(AF_INET, x),
            'decode': lambda x: inet_ntop(AF_INET, x),
        }

    class ip4list(option):
        policy = {
            'format': 'string',
            'encode': lambda x: ''.join([inet_pton(AF_INET, i) for i in x]),
            'decode': lambda x: [
                inet_ntop(AF_INET, x[i * 4 : i * 4 + 4])
                for i in range(len(x) // 4)
            ],
        }



'''
drr
+++

The qdisc doesn't accept any parameters, but the class
accepts `quantum` parameter::

    ip.tc('add', 'drr', interface, '1:')
    ip.tc('add-class', 'drr', interface, '1:10', quantum=1600)
    ip.tc('add-class', 'drr', interface, '1:20', quantum=1600)

'''
from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_ROOT
from pyroute2.netlink.rtnl.tcmsg.common import stats2 as c_stats2

parent = TC_H_ROOT


def get_class_parameters(kwarg):
    return {'attrs': [['TCA_DRR_QUANTUM', kwarg.get('quantum', 0)]]}


class options(nla):
    nla_map = (('TCA_DRR_UNSPEC', 'none'), ('TCA_DRR_QUANTUM', 'uint32'))


class stats(nla):
    fields = (('deficit', 'I'),)


class stats2(c_stats2):
    class stats_app(stats):
        pass



from pyroute2.common import map_namespace
from pyroute2.netlink import NLMSG_DONE
from pyroute2.netlink.nlsocket import Marshal

from . import ConnectorSocket, cn_msg

CN_IDX_PROC = 0x1

PROC_EVENT_NONE = 0x0
PROC_EVENT_FORK = 0x1
PROC_EVENT_EXEC = 0x2
PROC_EVENT_UID = 0x4
PROC_EVENT_GID = 0x40
PROC_EVENT_SID = 0x80
PROC_EVENT_PTRACE = 0x100
PROC_EVENT_COMM = 0x200
PROC_EVENT_COREDUMP = 0x40000000
PROC_EVENT_EXIT = 0x80000000

(PROC_BY_NAMES, PROC_BY_IDS) = map_namespace('PROC_', globals())

CN_IDX_PROC = 0x1
CN_VAL_PROC = 0x1

PROC_CN_MCAST_LISTEN = 0x1
PROC_CN_MCAST_IGNORE = 0x2


class proc_event_base(cn_msg):
    fields = cn_msg.fields + (
        ('what', 'I'),
        ('cpu', 'I'),
        ('timestamp_ns', 'Q'),
    )

    def decode(self):
        super().decode()
        self['event'] = PROC_BY_IDS.get(self['what'], 'UNDEFINED')


class proc_event_fork(proc_event_base):
    fields = proc_event_base.fields + (
        ('parent_pid', 'I'),
        ('parent_tgid', 'I'),
        ('child_pid', 'I'),
        ('child_tgid', 'I'),
    )


class proc_event_exec(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
    )


class proc_event_uid(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
        ('ruid', 'I'),
        ('rgid', 'I'),
    )


class proc_event_gid(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
        ('euid', 'I'),
        ('egid', 'I'),
    )


class proc_event_sid(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
    )


class proc_event_ptrace(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
        ('tracer_pid', 'I'),
        ('tracer_tgid', 'I'),
    )


class proc_event_comm(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
        ('comm', '16s'),
    )

    def decode(self):
        super().decode()
        self['comm'] = self['comm'].decode('utf-8').strip('\x00')


class proc_event_coredump(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
        ('parent_pid', 'I'),
        ('parent_tgid', 'I'),
    )


class proc_event_exit(proc_event_base):
    fields = proc_event_base.fields + (
        ('process_pid', 'I'),
        ('process_tgid', 'I'),
        ('exit_code', 'I'),
        ('exit_signal', 'I'),
        ('parent_pid', 'I'),
        ('parent_tgid', 'I'),
    )


class proc_event_control(cn_msg):
    fields = cn_msg.fields + (('action', 'I'),)


class ProcEventMarshal(Marshal):
    key_format = 'I'
    key_offset = 36
    error_type = -1
    msg_map = {
        PROC_EVENT_NONE: proc_event_base,
        PROC_EVENT_FORK: proc_event_fork,
        PROC_EVENT_EXEC: proc_event_exec,
        PROC_EVENT_UID: proc_event_uid,
        PROC_EVENT_GID: proc_event_gid,
        PROC_EVENT_SID: proc_event_sid,
        PROC_EVENT_PTRACE: proc_event_ptrace,
        PROC_EVENT_COMM: proc_event_comm,
        PROC_EVENT_COREDUMP: proc_event_coredump,
        PROC_EVENT_EXIT: proc_event_exit,
    }


class ProcEventSocket(ConnectorSocket):
    def __init__(self, fileno=None):
        super().__init__(fileno=fileno)
        self.marshal = ProcEventMarshal()

    def bind(self):
        return super().bind(groups=CN_IDX_PROC)

    def control(self, listen):
        msg = proc_event_control()
        msg['action'] = (
            PROC_CN_MCAST_LISTEN if listen else PROC_CN_MCAST_IGNORE
        )
        msg['idx'] = CN_IDX_PROC
        msg['val'] = CN_VAL_PROC
        msg['len'] = 4  # FIXME payload length calculation
        msg_type = NLMSG_DONE
        self.put(msg, msg_type, msg_flags=0, msg_seq=0)
        return tuple(self.get(msg_seq=-1))



'''
IPQ -- userspace firewall
=========================

Netlink family for dealing with `QUEUE` iptables
target. All the packets routed to the target
`QUEUE` should be handled by a userspace program
and the program should response with a verdict.
E.g., the verdict can be `NF_DROP` and in that
case the packet will be silently dropped, or
`NF_ACCEPT`, and the packet will be pass the
rule.
'''
from pyroute2.netlink import NLM_F_REQUEST, nlmsg
from pyroute2.netlink.nlsocket import Marshal, NetlinkSocket

# constants
IFNAMSIZ = 16
IPQ_MAX_PAYLOAD = 0x800

# IPQ messages
IPQM_BASE = 0x10
IPQM_MODE = IPQM_BASE + 1
IPQM_VERDICT = IPQM_BASE + 2
IPQM_PACKET = IPQM_BASE + 3

# IPQ modes
IPQ_COPY_NONE = 0
IPQ_COPY_META = 1
IPQ_COPY_PACKET = 2

# verdict types
NF_DROP = 0
NF_ACCEPT = 1
NF_STOLEN = 2
NF_QUEUE = 3
NF_REPEAT = 4
NF_STOP = 5


class ipq_base_msg(nlmsg):
    def decode(self):
        nlmsg.decode(self)
        self['payload'] = self.buf.read(self['data_len'])

    def encode(self):
        init = self.buf.tell()
        nlmsg.encode(self)
        if 'payload' in self:
            self.buf.write(self['payload'])
            self.update_length(init)


class ipq_packet_msg(ipq_base_msg):
    fields = (
        ('packet_id', 'L'),
        ('mark', 'L'),
        ('timestamp_sec', 'l'),
        ('timestamp_usec', 'l'),
        ('hook', 'I'),
        ('indev_name', '%is' % IFNAMSIZ),
        ('outdev_name', '%is' % IFNAMSIZ),
        ('hw_protocol', '>H'),
        ('hw_type', 'H'),
        ('hw_addrlen', 'B'),
        ('hw_addr', '6B'),
        ('__pad', '9x'),
        ('data_len', 'I'),
        ('__pad', '4x'),
    )


class ipq_mode_msg(nlmsg):
    pack = 'struct'
    fields = (
        ('value', 'B'),
        ('__pad', '7x'),
        ('range', 'I'),
        ('__pad', '12x'),
    )


class ipq_verdict_msg(ipq_base_msg):
    pack = 'struct'
    fields = (
        ('value', 'I'),
        ('__pad', '4x'),
        ('id', 'L'),
        ('data_len', 'I'),
        ('__pad', '4x'),
    )


class MarshalIPQ(Marshal):
    msg_map = {
        IPQM_MODE: ipq_mode_msg,
        IPQM_VERDICT: ipq_verdict_msg,
        IPQM_PACKET: ipq_packet_msg,
    }


class IPQSocket(NetlinkSocket):
    '''
    Low-level socket interface. Provides all the
    usual socket does, can be used in poll/select,
    doesn't create any implicit threads.
    '''

    def bind(self, mode=IPQ_COPY_PACKET):
        '''
        Bind the socket and performs IPQ mode configuration.
        The only parameter is mode, the default value is
        IPQ_COPY_PACKET (copy all the packet data).
        '''
        NetlinkSocket.bind(self, groups=0, pid=0)
        self.register_policy(MarshalIPQ.msg_map)
        msg = ipq_mode_msg()
        msg['value'] = mode
        msg['range'] = IPQ_MAX_PAYLOAD
        msg['header']['type'] = IPQM_MODE
        msg['header']['flags'] = NLM_F_REQUEST
        msg.encode()
        self.sendto(msg.data, (0, 0))

    def verdict(self, seq, v):
        '''
        Issue a verdict `v` for a packet `seq`.
        '''
        msg = ipq_verdict_msg()
        msg['value'] = v
        msg['id'] = seq
        msg['data_len'] = 0
        msg['header']['type'] = IPQM_VERDICT
        msg['header']['flags'] = NLM_F_REQUEST
        msg.encode()
        self.sendto(msg.buf.getvalue(), (0, 0))



'''
devlink module
==============
'''
from pyroute2.common import map_namespace
from pyroute2.netlink import genlmsg, nla
from pyroute2.netlink.generic import GenericNetlinkSocket
from pyroute2.netlink.nlsocket import Marshal

# devlink commands
DEVLINK_CMD_UNSPEC = 0
DEVLINK_CMD_GET = 1
DEVLINK_CMD_SET = 2
DEVLINK_CMD_NEW = 3
DEVLINK_CMD_DEL = 4
DEVLINK_CMD_PORT_GET = 5
DEVLINK_CMD_PORT_SET = 6
DEVLINK_CMD_PORT_NEW = 7
DEVLINK_CMD_PORT_DEL = 8
DEVLINK_CMD_PORT_SPLIT = 9
DEVLINK_CMD_PORT_UNSPLIT = 10
DEVLINK_CMD_SB_GET = 11
DEVLINK_CMD_SB_SET = 12
DEVLINK_CMD_SB_NEW = 13
DEVLINK_CMD_SB_DEL = 14
DEVLINK_CMD_SB_POOL_GET = 15
DEVLINK_CMD_SB_POOL_SET = 16
DEVLINK_CMD_SB_POOL_NEW = 17
DEVLINK_CMD_SB_POOL_DEL = 18
DEVLINK_CMD_SB_PORT_POOL_GET = 19
DEVLINK_CMD_SB_PORT_POOL_SET = 20
DEVLINK_CMD_SB_PORT_POOL_NEW = 21
DEVLINK_CMD_SB_PORT_POOL_DEL = 22
DEVLINK_CMD_SB_TC_POOL_BIND_GET = 23
DEVLINK_CMD_SB_TC_POOL_BIND_SET = 24
DEVLINK_CMD_SB_TC_POOL_BIND_NEW = 25
DEVLINK_CMD_SB_TC_POOL_BIND_DEL = 26
DEVLINK_CMD_SB_OCC_SNAPSHOT = 27
DEVLINK_CMD_SB_OCC_MAX_CLEAR = 28
DEVLINK_CMD_ESWITCH_GET = 29
DEVLINK_CMD_ESWITCH_SET = 30
DEVLINK_CMD_DPIPE_TABLE_GET = 31
DEVLINK_CMD_DPIPE_ENTRIES_GET = 32
DEVLINK_CMD_DPIPE_HEADERS_GET = 33
DEVLINK_CMD_DPIPE_TABLE_COUNTERS_SET = 34
DEVLINK_CMD_RESOURCE_SET = 35
DEVLINK_CMD_RESOURCE_DUMP = 36
DEVLINK_CMD_RELOAD = 37
DEVLINK_CMD_PARAM_GET = 38
DEVLINK_CMD_PARAM_SET = 39
DEVLINK_CMD_PARAM_NEW = 40
DEVLINK_CMD_PARAM_DEL = 41
DEVLINK_CMD_REGION_GET = 42
DEVLINK_CMD_REGION_SET = 43
DEVLINK_CMD_REGION_NEW = 44
DEVLINK_CMD_REGION_DEL = 45
DEVLINK_CMD_REGION_READ = 46
DEVLINK_CMD_PORT_PARAM_GET = 47
DEVLINK_CMD_PORT_PARAM_SET = 48
DEVLINK_CMD_PORT_PARAM_NEW = 49
DEVLINK_CMD_PORT_PARAM_DEL = 50
DEVLINK_CMD_INFO_GET = 51
DEVLINK_CMD_HEALTH_REPORTER_GET = 52
DEVLINK_CMD_HEALTH_REPORTER_SET = 53
DEVLINK_CMD_HEALTH_REPORTER_RECOVER = 54
DEVLINK_CMD_HEALTH_REPORTER_DIAGNOSE = 55
DEVLINK_CMD_HEALTH_REPORTER_DUMP_GET = 56
DEVLINK_CMD_HEALTH_REPORTER_DUMP_CLEAR = 57
DEVLINK_CMD_FLASH_UPDATE = 58
DEVLINK_CMD_FLASH_UPDATE_END = 59
DEVLINK_CMD_FLASH_UPDATE_STATUS = 60
DEVLINK_CMD_TRAP_GET = 61
DEVLINK_CMD_TRAP_SET = 62
DEVLINK_CMD_TRAP_NEW = 63
DEVLINK_CMD_TRAP_DEL = 64
DEVLINK_CMD_TRAP_GROUP_GET = 65
DEVLINK_CMD_TRAP_GROUP_SET = 66
DEVLINK_CMD_TRAP_GROUP_NEW = 67
DEVLINK_CMD_TRAP_GROUP_DEL = 68
DEVLINK_CMD_TRAP_POLICER_GET = 69
DEVLINK_CMD_TRAP_POLICER_SET = 70
DEVLINK_CMD_TRAP_POLICER_NEW = 71
DEVLINK_CMD_TRAP_POLICER_DEL = 72
DEVLINK_CMD_MAX = DEVLINK_CMD_TRAP_POLICER_DEL

(DEVLINK_NAMES, DEVLINK_VALUES) = map_namespace('DEVLINK_CMD_', globals())

# port type
DEVLINK_PORT_TYPE_NOTSET = 0
DEVLINK_PORT_TYPE_AUTO = 1
DEVLINK_PORT_TYPE_ETH = 2
DEVLINK_PORT_TYPE_IB = 3

# threshold type
DEVLINK_SB_POOL_TYPE_INGRESS = 0
DEVLINK_SB_POOL_TYPE_EGRESS = 1

DEVLINK_SB_THRESHOLD_TO_ALPHA_MAX = 20


class devlinkcmd(genlmsg):
    prefix = 'DEVLINK_ATTR_'
    nla_map = (
        ('DEVLINK_ATTR_UNSPEC', 'none'),
        ('DEVLINK_ATTR_BUS_NAME', 'asciiz'),
        ('DEVLINK_ATTR_DEV_NAME', 'asciiz'),
        ('DEVLINK_ATTR_PORT_INDEX', 'uint32'),
        ('DEVLINK_ATTR_PORT_TYPE', 'uint16'),
        ('DEVLINK_ATTR_PORT_DESIRED_TYPE', 'uint16'),
        ('DEVLINK_ATTR_PORT_NETDEV_IFINDEX', 'uint32'),
        ('DEVLINK_ATTR_PORT_NETDEV_NAME', 'asciiz'),
        ('DEVLINK_ATTR_PORT_IBDEV_NAME', 'asciiz'),
        ('DEVLINK_ATTR_PORT_SPLIT_COUNT', 'uint32'),
        ('DEVLINK_ATTR_PORT_SPLIT_GROUP', 'uint32'),
        ('DEVLINK_ATTR_SB_INDEX', 'uint32'),
        ('DEVLINK_ATTR_SB_SIZE', 'uint32'),
        ('DEVLINK_ATTR_SB_INGRESS_POOL_COUNT', 'uint16'),
        ('DEVLINK_ATTR_SB_EGRESS_POOL_COUNT', 'uint16'),
        ('DEVLINK_ATTR_SB_INGRESS_TC_COUNT', 'uint16'),
        ('DEVLINK_ATTR_SB_EGRESS_TC_COUNT', 'uint16'),
        ('DEVLINK_ATTR_SB_POOL_INDEX', 'uint16'),
        ('DEVLINK_ATTR_SB_POOL_TYPE', 'uint8'),
        ('DEVLINK_ATTR_SB_POOL_SIZE', 'uint32'),
        ('DEVLINK_ATTR_SB_POOL_THRESHOLD_TYPE', 'uint8'),
        ('DEVLINK_ATTR_SB_THRESHOLD', 'uint32'),
        ('DEVLINK_ATTR_SB_TC_INDEX', 'uint16'),
        ('DEVLINK_ATTR_SB_OCC_CUR', 'uint32'),
        ('DEVLINK_ATTR_SB_OCC_MAX', 'uint32'),
        ('DEVLINK_ATTR_ESWITCH_MODE', 'uint16'),
        ('DEVLINK_ATTR_ESWITCH_INLINE_MODE', 'uint8'),
        ('DEVLINK_ATTR_DPIPE_TABLES', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_TABLE', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_TABLE_NAME', 'asciiz'),
        ('DEVLINK_ATTR_DPIPE_TABLE_SIZE', 'uint64'),
        ('DEVLINK_ATTR_DPIPE_TABLE_MATCHES', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_TABLE_ACTIONS', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_TABLE_COUNTERS_ENABLED', 'uint8'),
        ('DEVLINK_ATTR_DPIPE_ENTRIES', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_ENTRY', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_ENTRY_INDEX', 'uint64'),
        ('DEVLINK_ATTR_DPIPE_ENTRY_MATCH_VALUES', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_ENTRY_ACTION_VALUES', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_ENTRY_COUNTER', 'uint64'),
        ('DEVLINK_ATTR_DPIPE_MATCH', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_MATCH_VALUE', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_MATCH_TYPE', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_ACTION', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_ACTION_VALUE', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_ACTION_TYPE', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_VALUE', 'none'),
        ('DEVLINK_ATTR_DPIPE_VALUE_MASK', 'none'),
        ('DEVLINK_ATTR_DPIPE_VALUE_MAPPING', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_HEADERS', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_HEADER', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_HEADER_NAME', 'asciiz'),
        ('DEVLINK_ATTR_DPIPE_HEADER_ID', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_HEADER_FIELDS', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_HEADER_GLOBAL', 'uint8'),
        ('DEVLINK_ATTR_DPIPE_HEADER_INDEX', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_FIELD', 'devlink'),
        ('DEVLINK_ATTR_DPIPE_FIELD_NAME', 'asciiz'),
        ('DEVLINK_ATTR_DPIPE_FIELD_ID', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_FIELD_BITWIDTH', 'uint32'),
        ('DEVLINK_ATTR_DPIPE_FIELD_MAPPING_TYPE', 'uint32'),
        ('DEVLINK_ATTR_PAD', 'none'),
        ('DEVLINK_ATTR_ESWITCH_ENCAP_MODE', 'uint8'),
        ('DEVLINK_ATTR_RESOURCE_LIST', 'devlink'),
        ('DEVLINK_ATTR_RESOURCE', 'devlink'),
        ('DEVLINK_ATTR_RESOURCE_NAME', 'asciiz'),
        ('DEVLINK_ATTR_RESOURCE_ID', 'uint64'),
        ('DEVLINK_ATTR_RESOURCE_SIZE', 'uint64'),
        ('DEVLINK_ATTR_RESOURCE_SIZE_NEW', 'uint64'),
        ('DEVLINK_ATTR_RESOURCE_SIZE_VALID', 'uint8'),
        ('DEVLINK_ATTR_RESOURCE_SIZE_MIN', 'uint64'),
        ('DEVLINK_ATTR_RESOURCE_SIZE_MAX', 'uint64'),
        ('DEVLINK_ATTR_RESOURCE_SIZE_GRAN', 'uint64'),
        ('DEVLINK_ATTR_RESOURCE_UNIT', 'uint8'),
        ('DEVLINK_ATTR_RESOURCE_OCC', 'uint64'),
        ('DEVLINK_ATTR_DPIPE_TABLE_RESOURCE_ID', 'uint64'),
        ('DEVLINK_ATTR_DPIPE_TABLE_RESOURCE_UNITS', 'uint64'),
        ('DEVLINK_ATTR_PORT_FLAVOUR', 'uint16'),
        ('DEVLINK_ATTR_PORT_NUMBER', 'uint32'),
        ('DEVLINK_ATTR_PORT_SPLIT_SUBPORT_NUMBER', 'uint32'),
        ('DEVLINK_ATTR_PARAM', 'devlink'),
        ('DEVLINK_ATTR_PARAM_NAME', 'asciiz'),
        ('DEVLINK_ATTR_PARAM_GENERIC', 'flag'),
        ('DEVLINK_ATTR_PARAM_TYPE', 'uint8'),
        ('DEVLINK_ATTR_PARAM_VALUES_LIST', 'devlink'),
        ('DEVLINK_ATTR_PARAM_VALUE', 'devlink'),
        ('DEVLINK_ATTR_PARAM_VALUE_DATA', 'none'),
        ('DEVLINK_ATTR_PARAM_VALUE_CMODE', 'uint8'),
        ('DEVLINK_ATTR_REGION_NAME', 'asciiz'),
        ('DEVLINK_ATTR_REGION_SIZE', 'uint64'),
        ('DEVLINK_ATTR_REGION_SNAPSHOTS', 'devlink'),
        ('DEVLINK_ATTR_REGION_SNAPSHOT', 'devlink'),
        ('DEVLINK_ATTR_REGION_SNAPSHOT_ID', 'uint32'),
        ('DEVLINK_ATTR_REGION_CHUNKS', 'devlink'),
        ('DEVLINK_ATTR_REGION_CHUNK', 'devlink'),
        ('DEVLINK_ATTR_REGION_CHUNK_DATA', 'binary'),
        ('DEVLINK_ATTR_REGION_CHUNK_ADDR', 'uint64'),
        ('DEVLINK_ATTR_REGION_CHUNK_LEN', 'uint64'),
        ('DEVLINK_ATTR_INFO_DRIVER_NAME', 'asciiz'),
        ('DEVLINK_ATTR_INFO_SERIAL_NUMBER', 'asciiz'),
        ('DEVLINK_ATTR_INFO_VERSION_FIXED', 'devlink'),
        ('DEVLINK_ATTR_INFO_VERSION_RUNNING', 'devlink'),
        ('DEVLINK_ATTR_INFO_VERSION_STORED', 'devlink'),
        ('DEVLINK_ATTR_INFO_VERSION_NAME', 'asciiz'),
        ('DEVLINK_ATTR_INFO_VERSION_VALUE', 'asciiz'),
        ('DEVLINK_ATTR_SB_POOL_CELL_SIZE', 'uint32'),
        ('DEVLINK_ATTR_FMSG', 'devlink'),
        ('DEVLINK_ATTR_FMSG_OBJ_NEST_START', 'flag'),
        ('DEVLINK_ATTR_FMSG_PAIR_NEST_START', 'flag'),
        ('DEVLINK_ATTR_FMSG_ARR_NEST_START', 'flag'),
        ('DEVLINK_ATTR_FMSG_NEST_END', 'flag'),
        ('DEVLINK_ATTR_FMSG_OBJ_NAME', 'asciiz'),
        ('DEVLINK_ATTR_FMSG_OBJ_VALUE_TYPE', 'uint8'),
        ('DEVLINK_ATTR_FMSG_OBJ_VALUE_DATA', 'none'),
        ('DEVLINK_ATTR_HEALTH_REPORTER', 'devlink'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_NAME', 'asciiz'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_STATE', 'uint8'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_ERR_COUNT', 'uint64'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_RECOVER_COUNT', 'uint64'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_DUMP_TS', 'uint64'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_GRACEFUL_PERIOD', 'uint64'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_AUTO_RECOVER', 'uint8'),
        ('DEVLINK_ATTR_FLASH_UPDATE_FILE_NAME', 'asciiz'),
        ('DEVLINK_ATTR_FLASH_UPDATE_COMPONENT', 'asciiz'),
        ('DEVLINK_ATTR_FLASH_UPDATE_STATUS_MSG', 'asciiz'),
        ('DEVLINK_ATTR_FLASH_UPDATE_STATUS_DONE', 'uint64'),
        ('DEVLINK_ATTR_FLASH_UPDATE_STATUS_TOTAL', 'uint64'),
        ('DEVLINK_ATTR_PORT_PCI_PF_NUMBER', 'uint16'),
        ('DEVLINK_ATTR_PORT_PCI_VF_NUMBER', 'uint16'),
        ('DEVLINK_ATTR_STATS', 'devlink'),
        ('DEVLINK_ATTR_TRAP_NAME', 'asciiz'),
        ('DEVLINK_ATTR_TRAP_ACTION', 'uint8'),
        ('DEVLINK_ATTR_TRAP_TYPE', 'uint8'),
        ('DEVLINK_ATTR_TRAP_GENERIC', 'flag'),
        ('DEVLINK_ATTR_TRAP_METADATA', 'devlink'),
        ('DEVLINK_ATTR_TRAP_GROUP_NAME', 'asciiz'),
        ('DEVLINK_ATTR_RELOAD_FAILED', 'uint8'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_DUMP_TS_NS', 'uint64'),
        ('DEVLINK_ATTR_NETNS_FD', 'uint32'),
        ('DEVLINK_ATTR_NETNS_PID', 'uint32'),
        ('DEVLINK_ATTR_NETNS_ID', 'uint32'),
        ('DEVLINK_ATTR_HEALTH_REPORTER_AUTO_DUMP', 'uint8'),
        ('DEVLINK_ATTR_TRAP_POLICER_ID', 'uint32'),
        ('DEVLINK_ATTR_TRAP_POLICER_RATE', 'uint64'),
        ('DEVLINK_ATTR_TRAP_POLICER_BURST', 'uint64'),
        ('DEVLINK_ATTR_PORT_FUNCTION', 'devlink'),
        ('DEVLINK_ATTR_INFO_BOARD_SERIAL_NUMBER', 'asciiz'),
        ('DEVLINK_ATTR_PORT_LANES', 'uint32'),
        ('DEVLINK_ATTR_PORT_SPLITTABLE', 'uint8'),
    )

    class devlink(nla):
        prefix = 'DEVLINK_ATTR_'
        nla_map = (
            ('DEVLINK_ATTR_UNSPEC', 'none'),
            ('DEVLINK_ATTR_BUS_NAME', 'asciiz'),
            ('DEVLINK_ATTR_DEV_NAME', 'asciiz'),
            ('DEVLINK_ATTR_PORT_INDEX', 'uint32'),
            ('DEVLINK_ATTR_PORT_TYPE', 'uint16'),
            ('DEVLINK_ATTR_PORT_DESIRED_TYPE', 'uint16'),
            ('DEVLINK_ATTR_PORT_NETDEV_IFINDEX', 'uint32'),
            ('DEVLINK_ATTR_PORT_NETDEV_NAME', 'asciiz'),
            ('DEVLINK_ATTR_PORT_IBDEV_NAME', 'asciiz'),
            ('DEVLINK_ATTR_PORT_SPLIT_COUNT', 'uint32'),
            ('DEVLINK_ATTR_PORT_SPLIT_GROUP', 'uint32'),
            ('DEVLINK_ATTR_SB_INDEX', 'uint32'),
            ('DEVLINK_ATTR_SB_SIZE', 'uint32'),
            ('DEVLINK_ATTR_SB_INGRESS_POOL_COUNT', 'uint16'),
            ('DEVLINK_ATTR_SB_EGRESS_POOL_COUNT', 'uint16'),
            ('DEVLINK_ATTR_SB_INGRESS_TC_COUNT', 'uint16'),
            ('DEVLINK_ATTR_SB_EGRESS_TC_COUNT', 'uint16'),
            ('DEVLINK_ATTR_SB_POOL_INDEX', 'uint16'),
            ('DEVLINK_ATTR_SB_POOL_TYPE', 'uint8'),
            ('DEVLINK_ATTR_SB_POOL_SIZE', 'uint32'),
            ('DEVLINK_ATTR_SB_POOL_THRESHOLD_TYPE', 'uint8'),
            ('DEVLINK_ATTR_SB_THRESHOLD', 'uint32'),
            ('DEVLINK_ATTR_SB_TC_INDEX', 'uint16'),
            ('DEVLINK_ATTR_SB_OCC_CUR', 'uint32'),
            ('DEVLINK_ATTR_SB_OCC_MAX', 'uint32'),
            ('DEVLINK_ATTR_ESWITCH_MODE', 'uint16'),
            ('DEVLINK_ATTR_ESWITCH_INLINE_MODE', 'uint8'),
            ('DEVLINK_ATTR_DPIPE_TABLES', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_TABLE', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_TABLE_NAME', 'asciiz'),
            ('DEVLINK_ATTR_DPIPE_TABLE_SIZE', 'uint64'),
            ('DEVLINK_ATTR_DPIPE_TABLE_MATCHES', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_TABLE_ACTIONS', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_TABLE_COUNTERS_ENABLED', 'uint8'),
            ('DEVLINK_ATTR_DPIPE_ENTRIES', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_ENTRY', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_ENTRY_INDEX', 'uint64'),
            ('DEVLINK_ATTR_DPIPE_ENTRY_MATCH_VALUES', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_ENTRY_ACTION_VALUES', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_ENTRY_COUNTER', 'uint64'),
            ('DEVLINK_ATTR_DPIPE_MATCH', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_MATCH_VALUE', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_MATCH_TYPE', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_ACTION', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_ACTION_VALUE', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_ACTION_TYPE', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_VALUE', 'none'),
            ('DEVLINK_ATTR_DPIPE_VALUE_MASK', 'none'),
            ('DEVLINK_ATTR_DPIPE_VALUE_MAPPING', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_HEADERS', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_HEADER', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_HEADER_NAME', 'asciiz'),
            ('DEVLINK_ATTR_DPIPE_HEADER_ID', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_HEADER_FIELDS', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_HEADER_GLOBAL', 'uint8'),
            ('DEVLINK_ATTR_DPIPE_HEADER_INDEX', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_FIELD', 'recursive'),
            ('DEVLINK_ATTR_DPIPE_FIELD_NAME', 'asciiz'),
            ('DEVLINK_ATTR_DPIPE_FIELD_ID', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_FIELD_BITWIDTH', 'uint32'),
            ('DEVLINK_ATTR_DPIPE_FIELD_MAPPING_TYPE', 'uint32'),
            ('DEVLINK_ATTR_PAD', 'none'),
            ('DEVLINK_ATTR_ESWITCH_ENCAP_MODE', 'uint8'),
            ('DEVLINK_ATTR_RESOURCE_LIST', 'recursive'),
            ('DEVLINK_ATTR_RESOURCE', 'recursive'),
            ('DEVLINK_ATTR_RESOURCE_NAME', 'asciiz'),
            ('DEVLINK_ATTR_RESOURCE_ID', 'uint64'),
            ('DEVLINK_ATTR_RESOURCE_SIZE', 'uint64'),
            ('DEVLINK_ATTR_RESOURCE_SIZE_NEW', 'uint64'),
            ('DEVLINK_ATTR_RESOURCE_SIZE_VALID', 'uint8'),
            ('DEVLINK_ATTR_RESOURCE_SIZE_MIN', 'uint64'),
            ('DEVLINK_ATTR_RESOURCE_SIZE_MAX', 'uint64'),
            ('DEVLINK_ATTR_RESOURCE_SIZE_GRAN', 'uint64'),
            ('DEVLINK_ATTR_RESOURCE_UNIT', 'uint8'),
            ('DEVLINK_ATTR_RESOURCE_OCC', 'uint64'),
            ('DEVLINK_ATTR_DPIPE_TABLE_RESOURCE_ID', 'uint64'),
            ('DEVLINK_ATTR_DPIPE_TABLE_RESOURCE_UNITS', 'uint64'),
            ('DEVLINK_ATTR_PORT_FLAVOUR', 'uint16'),
            ('DEVLINK_ATTR_PORT_NUMBER', 'uint32'),
            ('DEVLINK_ATTR_PORT_SPLIT_SUBPORT_NUMBER', 'uint32'),
            ('DEVLINK_ATTR_PARAM', 'recursive'),
            ('DEVLINK_ATTR_PARAM_NAME', 'asciiz'),
            ('DEVLINK_ATTR_PARAM_GENERIC', 'flag'),
            ('DEVLINK_ATTR_PARAM_TYPE', 'uint8'),
            ('DEVLINK_ATTR_PARAM_VALUES_LIST', 'recursive'),
            ('DEVLINK_ATTR_PARAM_VALUE', 'recursive'),
            ('DEVLINK_ATTR_PARAM_VALUE_DATA', 'none'),
            ('DEVLINK_ATTR_PARAM_VALUE_CMODE', 'uint8'),
            ('DEVLINK_ATTR_REGION_NAME', 'asciiz'),
            ('DEVLINK_ATTR_REGION_SIZE', 'uint64'),
            ('DEVLINK_ATTR_REGION_SNAPSHOTS', 'recursive'),
            ('DEVLINK_ATTR_REGION_SNAPSHOT', 'recursive'),
            ('DEVLINK_ATTR_REGION_SNAPSHOT_ID', 'uint32'),
            ('DEVLINK_ATTR_REGION_CHUNKS', 'recursive'),
            ('DEVLINK_ATTR_REGION_CHUNK', 'recursive'),
            ('DEVLINK_ATTR_REGION_CHUNK_DATA', 'binary'),
            ('DEVLINK_ATTR_REGION_CHUNK_ADDR', 'uint64'),
            ('DEVLINK_ATTR_REGION_CHUNK_LEN', 'uint64'),
            ('DEVLINK_ATTR_INFO_DRIVER_NAME', 'asciiz'),
            ('DEVLINK_ATTR_INFO_SERIAL_NUMBER', 'asciiz'),
            ('DEVLINK_ATTR_INFO_VERSION_FIXED', 'recursive'),
            ('DEVLINK_ATTR_INFO_VERSION_RUNNING', 'recursive'),
            ('DEVLINK_ATTR_INFO_VERSION_STORED', 'recursive'),
            ('DEVLINK_ATTR_INFO_VERSION_NAME', 'asciiz'),
            ('DEVLINK_ATTR_INFO_VERSION_VALUE', 'asciiz'),
            ('DEVLINK_ATTR_SB_POOL_CELL_SIZE', 'uint32'),
            ('DEVLINK_ATTR_FMSG', 'recursive'),
            ('DEVLINK_ATTR_FMSG_OBJ_NEST_START', 'flag'),
            ('DEVLINK_ATTR_FMSG_PAIR_NEST_START', 'flag'),
            ('DEVLINK_ATTR_FMSG_ARR_NEST_START', 'flag'),
            ('DEVLINK_ATTR_FMSG_NEST_END', 'flag'),
            ('DEVLINK_ATTR_FMSG_OBJ_NAME', 'asciiz'),
            ('DEVLINK_ATTR_FMSG_OBJ_VALUE_TYPE', 'uint8'),
            ('DEVLINK_ATTR_FMSG_OBJ_VALUE_DATA', 'none'),
            ('DEVLINK_ATTR_HEALTH_REPORTER', 'recursive'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_NAME', 'asciiz'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_STATE', 'uint8'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_ERR_COUNT', 'uint64'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_RECOVER_COUNT', 'uint64'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_DUMP_TS', 'uint64'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_GRACEFUL_PERIOD', 'uint64'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_AUTO_RECOVER', 'uint8'),
            ('DEVLINK_ATTR_FLASH_UPDATE_FILE_NAME', 'asciiz'),
            ('DEVLINK_ATTR_FLASH_UPDATE_COMPONENT', 'asciiz'),
            ('DEVLINK_ATTR_FLASH_UPDATE_STATUS_MSG', 'asciiz'),
            ('DEVLINK_ATTR_FLASH_UPDATE_STATUS_DONE', 'uint64'),
            ('DEVLINK_ATTR_FLASH_UPDATE_STATUS_TOTAL', 'uint64'),
            ('DEVLINK_ATTR_PORT_PCI_PF_NUMBER', 'uint16'),
            ('DEVLINK_ATTR_PORT_PCI_VF_NUMBER', 'uint16'),
            ('DEVLINK_ATTR_STATS', 'recursive'),
            ('DEVLINK_ATTR_TRAP_NAME', 'asciiz'),
            ('DEVLINK_ATTR_TRAP_ACTION', 'uint8'),
            ('DEVLINK_ATTR_TRAP_TYPE', 'uint8'),
            ('DEVLINK_ATTR_TRAP_GENERIC', 'flag'),
            ('DEVLINK_ATTR_TRAP_METADATA', 'recursive'),
            ('DEVLINK_ATTR_TRAP_GROUP_NAME', 'asciiz'),
            ('DEVLINK_ATTR_RELOAD_FAILED', 'uint8'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_DUMP_TS_NS', 'uint64'),
            ('DEVLINK_ATTR_NETNS_FD', 'uint32'),
            ('DEVLINK_ATTR_NETNS_PID', 'uint32'),
            ('DEVLINK_ATTR_NETNS_ID', 'uint32'),
            ('DEVLINK_ATTR_HEALTH_REPORTER_AUTO_DUMP', 'uint8'),
            ('DEVLINK_ATTR_TRAP_POLICER_ID', 'uint32'),
            ('DEVLINK_ATTR_TRAP_POLICER_RATE', 'uint64'),
            ('DEVLINK_ATTR_TRAP_POLICER_BURST', 'uint64'),
            ('DEVLINK_ATTR_PORT_FUNCTION', 'recursive'),
            ('DEVLINK_ATTR_INFO_BOARD_SERIAL_NUMBER', 'asciiz'),
            ('DEVLINK_ATTR_PORT_LANES', 'uint32'),
            ('DEVLINK_ATTR_PORT_SPLITTABLE', 'uint8'),
        )


class MarshalDevlink(Marshal):
    msg_map = {
        DEVLINK_CMD_UNSPEC: devlinkcmd,
        DEVLINK_CMD_GET: devlinkcmd,
        DEVLINK_CMD_SET: devlinkcmd,
        DEVLINK_CMD_NEW: devlinkcmd,
        DEVLINK_CMD_DEL: devlinkcmd,
        DEVLINK_CMD_PORT_GET: devlinkcmd,
        DEVLINK_CMD_PORT_SET: devlinkcmd,
        DEVLINK_CMD_PORT_NEW: devlinkcmd,
        DEVLINK_CMD_PORT_DEL: devlinkcmd,
        DEVLINK_CMD_PORT_SPLIT: devlinkcmd,
        DEVLINK_CMD_PORT_UNSPLIT: devlinkcmd,
        DEVLINK_CMD_SB_GET: devlinkcmd,
        DEVLINK_CMD_SB_SET: devlinkcmd,
        DEVLINK_CMD_SB_NEW: devlinkcmd,
        DEVLINK_CMD_SB_DEL: devlinkcmd,
        DEVLINK_CMD_SB_POOL_GET: devlinkcmd,
        DEVLINK_CMD_SB_POOL_SET: devlinkcmd,
        DEVLINK_CMD_SB_POOL_NEW: devlinkcmd,
        DEVLINK_CMD_SB_POOL_DEL: devlinkcmd,
        DEVLINK_CMD_SB_PORT_POOL_GET: devlinkcmd,
        DEVLINK_CMD_SB_PORT_POOL_SET: devlinkcmd,
        DEVLINK_CMD_SB_PORT_POOL_NEW: devlinkcmd,
        DEVLINK_CMD_SB_PORT_POOL_DEL: devlinkcmd,
        DEVLINK_CMD_SB_TC_POOL_BIND_GET: devlinkcmd,
        DEVLINK_CMD_SB_TC_POOL_BIND_SET: devlinkcmd,
        DEVLINK_CMD_SB_TC_POOL_BIND_NEW: devlinkcmd,
        DEVLINK_CMD_SB_TC_POOL_BIND_DEL: devlinkcmd,
        DEVLINK_CMD_SB_OCC_SNAPSHOT: devlinkcmd,
        DEVLINK_CMD_SB_OCC_MAX_CLEAR: devlinkcmd,
        DEVLINK_CMD_ESWITCH_GET: devlinkcmd,
        DEVLINK_CMD_ESWITCH_SET: devlinkcmd,
        DEVLINK_CMD_DPIPE_TABLE_GET: devlinkcmd,
        DEVLINK_CMD_DPIPE_ENTRIES_GET: devlinkcmd,
        DEVLINK_CMD_DPIPE_HEADERS_GET: devlinkcmd,
        DEVLINK_CMD_DPIPE_TABLE_COUNTERS_SET: devlinkcmd,
        DEVLINK_CMD_RESOURCE_SET: devlinkcmd,
        DEVLINK_CMD_RESOURCE_DUMP: devlinkcmd,
        DEVLINK_CMD_RELOAD: devlinkcmd,
        DEVLINK_CMD_PARAM_GET: devlinkcmd,
        DEVLINK_CMD_PARAM_SET: devlinkcmd,
        DEVLINK_CMD_PARAM_NEW: devlinkcmd,
        DEVLINK_CMD_PARAM_DEL: devlinkcmd,
        DEVLINK_CMD_REGION_GET: devlinkcmd,
        DEVLINK_CMD_REGION_SET: devlinkcmd,
        DEVLINK_CMD_REGION_NEW: devlinkcmd,
        DEVLINK_CMD_REGION_DEL: devlinkcmd,
        DEVLINK_CMD_REGION_READ: devlinkcmd,
        DEVLINK_CMD_PORT_PARAM_GET: devlinkcmd,
        DEVLINK_CMD_PORT_PARAM_SET: devlinkcmd,
        DEVLINK_CMD_PORT_PARAM_NEW: devlinkcmd,
        DEVLINK_CMD_PORT_PARAM_DEL: devlinkcmd,
        DEVLINK_CMD_INFO_GET: devlinkcmd,
        DEVLINK_CMD_HEALTH_REPORTER_GET: devlinkcmd,
        DEVLINK_CMD_HEALTH_REPORTER_SET: devlinkcmd,
        DEVLINK_CMD_HEALTH_REPORTER_RECOVER: devlinkcmd,
        DEVLINK_CMD_HEALTH_REPORTER_DIAGNOSE: devlinkcmd,
        DEVLINK_CMD_HEALTH_REPORTER_DUMP_GET: devlinkcmd,
        DEVLINK_CMD_HEALTH_REPORTER_DUMP_CLEAR: devlinkcmd,
        DEVLINK_CMD_FLASH_UPDATE: devlinkcmd,
        DEVLINK_CMD_FLASH_UPDATE_END: devlinkcmd,
        DEVLINK_CMD_FLASH_UPDATE_STATUS: devlinkcmd,
        DEVLINK_CMD_TRAP_GET: devlinkcmd,
        DEVLINK_CMD_TRAP_SET: devlinkcmd,
        DEVLINK_CMD_TRAP_NEW: devlinkcmd,
        DEVLINK_CMD_TRAP_DEL: devlinkcmd,
        DEVLINK_CMD_TRAP_GROUP_GET: devlinkcmd,
        DEVLINK_CMD_TRAP_GROUP_SET: devlinkcmd,
        DEVLINK_CMD_TRAP_GROUP_NEW: devlinkcmd,
        DEVLINK_CMD_TRAP_GROUP_DEL: devlinkcmd,
        DEVLINK_CMD_TRAP_POLICER_GET: devlinkcmd,
        DEVLINK_CMD_TRAP_POLICER_SET: devlinkcmd,
        DEVLINK_CMD_TRAP_POLICER_NEW: devlinkcmd,
        DEVLINK_CMD_TRAP_POLICER_DEL: devlinkcmd,
    }

    def fix_message(self, msg):
        try:
            msg['event'] = DEVLINK_VALUES[msg['cmd']]
        except Exception:
            pass


class DevlinkSocket(GenericNetlinkSocket):
    def __init__(self):
        GenericNetlinkSocket.__init__(self)
        self.marshal = MarshalDevlink()

    def bind(self, groups=0, **kwarg):
        GenericNetlinkSocket.bind(
            self, 'devlink', devlinkcmd, groups, None, **kwarg
        )
